from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timezone
from pathlib import Path
from typing import Dict, List

import pydantic
import yaml

try:
    from shared.util import download_from_url_using_requests
except ModuleNotFoundError:
    from truss.templates.shared.util import download_from_url_using_requests

LAZY_DATA_RESOLVER_PATH = Path("/bptr/bptr-manifest")
NUM_WORKERS = 4


class Resolution(pydantic.BaseModel):
    url: str
    expiration_timestamp: int


class BasetenPointer(pydantic.BaseModel):
    """Specification for lazy data resolution for download of large files, similar to Git LFS pointers"""

    resolution: Resolution
    uid: str
    file_name: str
    hashtype: str
    hash: str
    size: int


class BasetenPointerManifest(pydantic.BaseModel):
    pointers: List[BasetenPointer]


class LazyDataResolver:
    def __init__(self, data_dir: Path):
        self._data_dir: Path = data_dir
        self._bptr_resolution: Dict[str, str] = _read_bptr_resolution()
        self._resolution_done = False

    def fetch(self):
        if self._resolution_done:
            return

        with ThreadPoolExecutor(NUM_WORKERS) as executor:
            futures = {}
            for file_name, resolved_url in self._bptr_resolution.items():
                futures[file_name] = executor.submit(
                    download_from_url_using_requests,
                    resolved_url,
                    self._data_dir / file_name,
                )
            for file_name, future in futures.items():
                if not future:
                    raise RuntimeError(f"Download failure for file {file_name}")
        self._resolution_done = True


def _read_bptr_resolution() -> Dict[str, str]:
    if not LAZY_DATA_RESOLVER_PATH.is_file():
        return {}
    bptr_manifest = BasetenPointerManifest(
        **yaml.safe_load(LAZY_DATA_RESOLVER_PATH.read_text())
    )
    resolution_map = {}
    for bptr in bptr_manifest.pointers:
        if bptr.resolution.expiration_timestamp < int(
            datetime.now(timezone.utc).timestamp()
        ):
            raise RuntimeError("Baseten pointer lazy data resolution has expired")
        resolution_map[bptr.file_name] = bptr.resolution.url
    return resolution_map



import asyncio
import concurrent.futures
import fcntl
import hashlib
import json
import multiprocessing
import os
import signal
import socket
import subprocess
import threading
import time
from itertools import count
from pathlib import Path
from typing import Any, Dict, List, Optional

import briton_pb2
import briton_pb2_grpc
import grpc
from fastapi import HTTPException
from outlines.models.transformers import TransformerTokenizer
from outlines.processors.structured import JSONLogitsProcessor
from transformers import AutoTokenizer, PreTrainedTokenizerFast
from truss.config.trt_llm import TrussTRTLLMBuildConfiguration
from truss.constants import OPENAI_COMPATIBLE_TAG

BRITON_PORT = 50051

MODEL_INPUT_TO_BRITON_FIELD = {
    "max_tokens": "request_output_len",
    "beam_width": "beam_width",
    "repetition_penalty": "repetition_penalty",
    "presence_penalty": "presence_penalty",
    "temperature": "temperature",
    "length_penalty": "len_penalty",
    "end_id": "end_id",
    "pad_id": "pad_id",
    "runtime_top_k": "runtime_top_k",
    "runtime_top_p": "runtime_top_p",
}

# Use a directory that can be picked up by baseten-fs
FSM_CACHE_DIR = "/cache/model/fsm_cache"

TOOL_CALL_IDS = {
    "llama": 128010,
    "mistral": 5,
}

TOOL_CALL_TOKENS = {
    "llama": "<|python_tag|>",
    "mistral": "[TOOL_CALLS]",
}


def is_port_available(port, host="localhost"):
    try:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            # Try to connect to the specified host and port
            s.bind((host, port))
            return True
    except OSError:
        # Port is not available
        return False


def briton_monitor(briton_process):
    while True:
        if briton_process.poll() is not None:
            print(
                f"Briton process has exited with code {briton_process.returncode}, exiting truss server"
            )
            pid = os.getpid()
            os.kill(pid, signal.SIGKILL)
        time.sleep(1)


class Engine:
    def __init__(self, **kwargs):
        self._loaded = False
        self._model = None
        self._config = kwargs["config"]
        self._data_dir = kwargs["data_dir"]
        self._stub = None
        self._secrets = kwargs["secrets"]
        self._request_id_counter = count(start=1)
        self._briton_process = None
        if OPENAI_COMPATIBLE_TAG not in self._config.get("model_metadata", {}):
            print(
                "Warning: Defaulting to openai compatible interface, even though not indicated in config."
            )

        if "trt_llm" not in self._config:
            raise ValueError("trt_llm config is required for this model")

        trtllm_config = self._config.get("trt_llm")
        truss_trtllm_build_config = TrussTRTLLMBuildConfiguration(
            **trtllm_config.get("build")
        )
        self._base_model = truss_trtllm_build_config.base_model
        self._tp_count = truss_trtllm_build_config.tensor_parallel_count
        self._tokenizer_repository = (
            truss_trtllm_build_config.checkpoint_repository.repo
        )
        self._kv_cache_free_gpu_mem_fraction = (
            truss_trtllm_build_config.kv_cache_free_gpu_mem_fraction
        )
        self._enable_kv_cache_reuse = (
            truss_trtllm_build_config.plugin_configuration.use_paged_context_fmha
        )

        self._hf_token = None
        try:
            self._hf_token = self._secrets.get("hf_access_token", None)
        except:  # noqa
            pass

        self._max_input_len = truss_trtllm_build_config.max_input_len
        self._max_beam_width = truss_trtllm_build_config.max_beam_width

        # TODO(@bdubayah): configure this based on CPU. But os.cpu_count() returns the
        # number of CPUs for the entire node, not just the container.
        self._max_fsm_workers = 10
        print(f"Using {self._max_fsm_workers} workers for FSM schema generation")

    def load(self):
        if self._loaded:
            return

        self._tokenizer = AutoTokenizer.from_pretrained(
            self._tokenizer_repository, token=self._hf_token
        )

        self._fsm_cache = FsmCache(
            Path(FSM_CACHE_DIR), self._tokenizer, self._max_fsm_workers
        )

        # We only support Llama and mistral with Briton, for which this should
        # apply.
        assert isinstance(self._tokenizer, PreTrainedTokenizerFast)

        # These are tokens outside of tokenizer.json. We need to pass these to
        # Briton, to pass to rust tokenizer.
        added_token_decoders = self._tokenizer.added_tokens_decoder
        added_tokens = [token for token in added_token_decoders.values()]

        self._saved_tokenizer_dir = str(self._data_dir / "saved_tokenizer")
        self._tokenizer.save_pretrained(self._saved_tokenizer_dir)

        # Pass tokenizer file to Briton for the rust tokenizer.
        tokenizer_file = Path(self._saved_tokenizer_dir) / "tokenizer.json"
        config_str = f"""
engine_path: "{self._data_dir.resolve()}"
hf_tokenizer: "{tokenizer_file.resolve()}"
kv_cache_free_gpu_mem_fraction: {self._kv_cache_free_gpu_mem_fraction}
enable_kv_cache_reuse: {"true" if self._enable_kv_cache_reuse else "false"}
fsm_cache_dir: "{FSM_CACHE_DIR}"
"""

        # Pass added tokens to Briton for the rust tokenizer.
        if len(added_tokens) > 0:
            config_str += "\n" + "\n".join(
                _serialize_added_tokens_to_config(added_tokens)
            )

        config_pbtxt_path = (self._data_dir / "briton_config.pbtxt").resolve()
        config_pbtxt_path.write_text(config_str)
        briton_env = os.environ.copy()
        if self._hf_token is not None:
            briton_env["HF_ACCESS_TOKEN"] = self._hf_token
        if self._tp_count is None or self._tp_count == 1:
            self._briton_process = subprocess.Popen(
                ["Briton", "--config", str(config_pbtxt_path)], env=briton_env
            )
        else:
            self._briton_process = subprocess.Popen(
                [
                    "mpirun",
                    "--allow-run-as-root",
                    "-n",
                    f"{self._tp_count}",
                    "Briton",
                    "--config",
                    str(config_pbtxt_path),
                ],
                env=briton_env,
            )
        while is_port_available(BRITON_PORT):
            print("Waiting for Briton to start")
            time.sleep(1)

        briton_monitor_thread = threading.Thread(
            target=briton_monitor, args=(self._briton_process,)
        )
        briton_monitor_thread.start()
        self._loaded = True

    def validate_input(self, model_input):
        beam_width = model_input.get("beam_width", None)
        # Beam width == 1.
        # There's no need to check if streaming is passed in the input.
        # Briton explicitly sets streaming to true in britonToTbRequest().
        # https://github.com/basetenlabs/baseten/blob/1c2c9cbe1adafc0c736566bd012abbe7d7e2c2da/briton/src/briton.cpp#L272
        if beam_width is not None and beam_width != 1:
            raise HTTPException(
                status_code=400, detail="TensorRT-LLM requires beam_width to equal 1"
            )

        # If Beam width != max_beam_width, TensorRt-LLM will fail an assert.
        # Since Briton sets streaming, the max_beam_width must aslo equal 1.
        if self._max_beam_width != 1:
            raise HTTPException(
                status_code=400,
                detail="TensorRT-LLM requires max_beam_width to equal 1.",
            )

    async def predict(self, model_input):
        """
        Run inference

        Note that the async nature of this function is a little tricky. Care is
        needed to make sure this function is a regular async function and not an
        async generator, i.e. there shouldn't be any direct yields in this
        function. This is because we need to support both streaming and
        non-streaming cases in this function. We do this by either returning an
        async-generator for the streaming case, or directly the full text for
        the other case. Returning an async generator for non-streaming case
        interferes with the open ai client proxy.
        """
        if self._stub is None:
            channel = grpc.aio.insecure_channel(f"localhost:{BRITON_PORT}")
            self._stub = briton_pb2_grpc.BritonStub(channel)

        # TODO(@bdubayah): refactor into smaller functions
        function_calling_schema = None
        tools = model_input.get("tools")
        tool_choice = model_input.get("tool_choice")
        force_tools = None
        if tool_choice is not None:
            if not (
                tool_choice in ["none", "required", "auto"]
                or isinstance(tool_choice, dict)
            ):
                raise HTTPException(
                    status_code=400,
                    detail="tool_choice must be 'none', 'required', 'auto', or an object of the form {'type': 'function', 'function': {'name': 'function_name'}}.",
                )
            if tool_choice == "none":
                tools = None
                tool_choice = None
            elif tool_choice == "required":
                if tools is None:
                    raise HTTPException(
                        status_code=400,
                        detail="tool_choice is 'required' but no tools provided.",
                    )
                force_tools = True
        if tools is not None:
            if model_input.get("response_format") is not None:
                raise HTTPException(
                    status_code=400,
                    detail="response_format is not allowed when tools are provided, unless tool_choice is 'none'.",
                )
            tool_schemas = {
                tool["function"]["name"]: create_tool_schema(tool) for tool in tools
            }
            if isinstance(tool_choice, dict):
                if tool_choice.get("type") != "function":
                    raise HTTPException(
                        status_code=400, detail="tool_choice['type'] must be function."
                    )
                if tool_choice.get("function") is None:
                    raise HTTPException(
                        status_code=400, detail="tool_choice['function'] required."
                    )
                if not isinstance(tool_choice["function"], dict):
                    raise HTTPException(
                        status_code=400,
                        detail="tool_choice['function'] must be an object.",
                    )
                if tool_choice["function"].get("name") is None:
                    raise HTTPException(
                        status_code=400,
                        detail="tool_choice['function']['name'] required.",
                    )
                if tool_choice["function"]["name"] not in tool_schemas:
                    raise HTTPException(
                        status_code=400,
                        detail=f"Tool choice function {tool_choice['function']['name']} not in tools.",
                    )
                tool_schemas = {
                    tool_choice["function"]["name"]: tool_schemas[
                        tool_choice["function"]["name"]
                    ]
                }
                force_tools = True
            elif tool_choice is None or tool_choice == "auto":
                force_tools = False
            function_calling_schema = {
                "type": "array",
                "items": {
                    "anyOf": list(tool_schemas.values()),
                },
                "minItems": 1,
            }

        prompt = model_input.get("prompt", None)
        if prompt is not None and tools is not None:
            raise HTTPException(
                status_code=400,
                detail="tools can only be provided in chat mode. Please set messages instead of prompt, remove tools, or set tool_choice to 'none'.",
            )
        if prompt is None and "messages" in model_input:
            messages = model_input.pop("messages")
            prompt = self._tokenizer.apply_chat_template(
                messages, tools=tools, tokenize=False, add_generation_prompt=True
            )
        if prompt is None or len(prompt) == 0:
            raise HTTPException(status_code=400, detail="Prompt cannot be empty.")

        self.validate_input(model_input)

        request_id = int(str(os.getpid()) + str(next(self._request_id_counter)))
        request = briton_pb2.InferenceRequest(
            request_id=request_id,
            input_text=prompt,
        )
        # Set default end_id and pad_id
        if (
            hasattr(self._tokenizer, "eos_token_id")
            and self._tokenizer.eos_token_id is not None
        ):
            request.end_id = self._tokenizer.eos_token_id
        if (
            hasattr(self._tokenizer, "pad_token_id")
            and self._tokenizer.pad_token_id is not None
        ):
            request.pad_id = self._tokenizer.pad_token_id
        # Add output schema hash if we're function calling or response_format is provided
        schema_hash = None
        try:
            schema_hash = (
                await self._fsm_cache.add_schema(function_calling_schema)
                if function_calling_schema is not None
                else await self._fsm_cache.add_schema_from_input(model_input)
            )
        # If the input schema is invalid, we should return a 400
        except NotImplementedError as ex:
            raise HTTPException(status_code=400, detail=str(ex))
        if schema_hash is not None:
            request.output_schema_hash = schema_hash
        if force_tools is not None:
            request.tools_id = TOOL_CALL_IDS[self._base_model]
            request.force_tools = force_tools
        set_briton_request_fields_from_model_input(model_input, request)
        for words in ["bad_words", "stop_words"]:
            if words in model_input:
                for word in model_input[words].split(","):
                    getattr(request, words).append(word)

        resp_iter = self._stub.Infer(request)

        async def generate():
            eos_token = (
                self._tokenizer.eos_token
                if hasattr(self._tokenizer, "eos_token")
                else None
            )
            tool_call_token = TOOL_CALL_TOKENS.get(self._base_model)
            async for response in resp_iter:
                output_text = response.output_text
                if tool_call_token:
                    output_text = output_text.removeprefix(tool_call_token)
                if eos_token:
                    output_text = output_text.removesuffix(eos_token)
                yield output_text

        async def build_response():
            eos_token = (
                self._tokenizer.eos_token
                if hasattr(self._tokenizer, "eos_token")
                else None
            )
            tool_call_token = TOOL_CALL_TOKENS.get(self._base_model)
            full_text = ""
            async for delta in resp_iter:
                full_text += delta.output_text
            if tool_call_token:
                full_text = full_text.removeprefix(tool_call_token)
            if eos_token:
                full_text = full_text.removesuffix(eos_token)
            return full_text

        try:
            if model_input.get("stream", True):
                gen = generate()
                first_chunk = await gen.__anext__()

                async def generate_after_first_chunk():
                    yield first_chunk
                    async for chunk in gen:
                        yield chunk

                return generate_after_first_chunk()
            else:
                return await build_response()
        except grpc.RpcError as ex:
            if (
                ex.code() == grpc.StatusCode.INVALID_ARGUMENT
                or ex.code() == grpc.StatusCode.UNIMPLEMENTED
            ):
                raise HTTPException(status_code=400, detail=ex.details())
            # If the error is another type of gRPC error, we should return a 500
            else:
                raise HTTPException(
                    status_code=500, detail=f"An error has occurred: {ex.details()}"
                )
        except Exception as ex:
            raise HTTPException(status_code=500, detail=f"An error has occurred: {ex}")


def _serialize_added_tokens_to_config(added_tokens: list) -> List[str]:
    """Serialize to pbtxt format."""
    lines = ["added_tokens {"]
    for added_token in added_tokens:
        token_lines = _serialize_added_token_to_config(added_token)
        lines.extend([f"  {line}" for line in token_lines])
    lines.append("}")
    return lines


def _serialize_added_token_to_config(added_token) -> List[str]:
    """Serialize to pbtxt format."""
    fields = [
        f'content: "{added_token.content}"',
        f"single_word: {added_token.single_word}",
        f"lstrip: {added_token.lstrip}",
        f"rstrip: {added_token.rstrip}",
        f"normalized: {added_token.normalized}",
        f"special: {added_token.special}",
    ]
    return [
        "tokens {",
        *[f"  {field}" for field in fields],
        "}",
    ]


def create_tool_schema(tool_json: Dict[str, Any]) -> Dict[str, Any]:
    return {
        "type": "object",
        "properties": {
            "name": {"const": tool_json["function"]["name"]},
            "parameters": tool_json["function"]["parameters"],
        },
        "required": ["name", "parameters"],
    }


outlines_tokenizer = None


def worker(vocab_size: int, end_id: int, schema: Dict[str, Any], output_path: Path):
    logits_processor = JSONLogitsProcessor(schema, outlines_tokenizer)
    guide = logits_processor.fsm
    states_to_tokens = {}
    for state, token_to_next_state in guide.states_to_token_maps.items():
        states_to_tokens[state] = briton_pb2.TokenToNextState(  # type: ignore[attr-defined]
            token_to_next_state=token_to_next_state
        )
    states_to_tokens_pb = briton_pb2.StatesToTokens(  # type: ignore[attr-defined]
        states_to_tokens=states_to_tokens,
        vocab_size=vocab_size,
        eos_token_id=end_id,
    )
    if not output_path.exists():
        try:
            # Open the file with flags to protect against concurrent writes.
            # O_CREAT: Create the file if it does not exist.
            # O_EXCL: Ensure that this call creates the file exclusively. If the file already exists, the call will fail.
            # O_WRONLY: Open the file for write-only access.
            fd = os.open(output_path, os.O_CREAT | os.O_EXCL | os.O_WRONLY)
            with os.fdopen(fd, "wb") as f:
                fcntl.flock(f, fcntl.LOCK_EX)
                f.write(states_to_tokens_pb.SerializeToString())
                fcntl.flock(f, fcntl.LOCK_UN)
        except FileExistsError:
            pass


def dummy_task():
    pass


class FsmCache:
    def __init__(self, cache_dir: Path, tokenizer: AutoTokenizer, max_workers: int):
        self._cache_dir = cache_dir
        if not self._cache_dir.exists():
            self._cache_dir.mkdir(parents=True, exist_ok=True)
        self._cache = set(f.name for f in self._cache_dir.iterdir() if f.is_file())
        self._lock = threading.Lock()
        self._tokenizer = tokenizer

        # Concurrent FSM generation initialization
        # Make sure we fork because (1) it's faster and (2) it seems that spawning
        # ends up being sequential
        multiprocessing.set_start_method("fork", force=True)
        global outlines_tokenizer
        outlines_tokenizer = TransformerTokenizer(tokenizer)
        # This is very important. The first time JSONLogitsProcessor is called, some library-wide
        # initializations are done in memory (that take 5s). By doing it before we fork, we avoid paying
        # that cost for each forked process.
        _ = JSONLogitsProcessor({"properties": {}}, outlines_tokenizer)
        self._executor = concurrent.futures.ProcessPoolExecutor(max_workers=max_workers)
        # We must create all processes BEFORE the GRPC python client is started to avoid errors
        # forking from the process GRPC is running in
        for _ in range(max_workers):
            self._executor.submit(dummy_task)

    async def add_schema(self, schema: Dict[str, Any]) -> str:
        schema_str = json.dumps(schema)
        schema_hash = hashlib.sha256(schema_str.encode()).hexdigest()
        if schema_hash not in self._cache:
            loop = asyncio.get_running_loop()
            await loop.run_in_executor(
                self._executor,
                worker,
                len(self._tokenizer.vocab),
                self._tokenizer.eos_token_id,
                schema,
                self._cache_dir / schema_hash,
            )
            with self._lock:
                self._cache.add(schema_hash)
        return schema_hash

    async def add_schema_from_input(self, model_input: Dict[str, Any]) -> Optional[str]:
        schema_hash = None
        schema = self._extract_schema(model_input)
        if schema is not None:
            schema_hash = await self.add_schema(schema)
        return schema_hash

    @staticmethod
    def _extract_schema(model_input: Dict[str, Any]) -> Optional[Dict[str, Any]]:
        if "response_format" not in model_input:
            return None
        response_format = model_input["response_format"]
        if "type" not in response_format or response_format["type"] != "json_schema":
            raise HTTPException(
                status_code=400,
                detail='response_format["type"] must be json_schema.',
            )
        if "json_schema" not in response_format:
            raise HTTPException(
                status_code=400,
                detail='response_format["json_schema"] must be provided.',
            )
        json_schema = response_format["json_schema"]
        if "schema" not in json_schema:
            raise HTTPException(
                status_code=400,
                detail='response_format["json_schema"]["schema"] must be provided.',
            )
        return json_schema["schema"]


def set_briton_request_fields_from_model_input(model_input, briton_request):
    for model_input_key, briton_field in MODEL_INPUT_TO_BRITON_FIELD.items():
        if model_input_key in model_input:
            model_input_value = model_input[model_input_key]
            setattr(briton_request, briton_field, model_input_value)



from pathlib import Path

import pytest
import requests

from truss.local.local_config_handler import LocalConfigHandler
from truss.tests.test_testing_utilities_for_other_tests import ensure_kill_all
from truss.truss_handle import TrussHandle


@pytest.mark.integration
def test_docker_server_truss():
    with ensure_kill_all():
        truss_root = Path(__file__).parent.parent.parent.resolve() / "truss"

        truss_dir = truss_root / "test_data" / "test_docker_server_truss"

        tr = TrussHandle(truss_dir)
        LocalConfigHandler.set_secret("hf_access_token", "123")
        _ = tr.docker_run(local_port=8090, detach=True, wait_for_server_ready=True)
        truss_server_addr = "http://localhost:8090"
        full_url = f"{truss_server_addr}/v1/models/model:predict"

        response = requests.post(full_url, json={})
        assert response.status_code == 200
        assert response.json() == {
            "message": "Hello World",
            "is_torch_cuda_available": False,
            "is_env_var_passed": True,
            "is_secret_mounted": True,
        }



import base64
import json
import os

import boto3
from boto3.s3.transfer import TransferConfig
from rich.progress import Progress


def base64_encoded_json_str(obj):
    return base64.b64encode(str.encode(json.dumps(obj))).decode("utf-8")


def multipart_upload_boto3(file_path, bucket_name, key, credentials):
    s3_resource = boto3.resource("s3", **credentials)
    filesize = os.stat(file_path).st_size

    # Create a new progress bar
    progress = Progress()

    # Add a new task to the progress bar
    task_id = progress.add_task("[cyan]Uploading...", total=filesize)

    with progress:

        def callback(bytes_transferred):
            # Update the progress bar
            progress.update(task_id, advance=bytes_transferred)

        s3_resource.Object(bucket_name, key).upload_file(
            file_path,
            Config=TransferConfig(
                max_concurrency=10,
                use_threads=True,
            ),
            Callback=callback,
        )



# This file contains shared code to be used in other tests
# TODO(pankaj): Using a tests file for shared code is not ideal, we should
# move it to a regular file. This is a short term hack.
import json
import shutil
import subprocess
import time
from contextlib import contextmanager

from truss.build import kill_all
from truss.constants import TRUSS
from truss.docker import get_containers

DISK_SPACE_LOW_PERCENTAGE = 20


@contextmanager
def ensure_kill_all():
    try:
        with _show_container_logs_if_raised():
            yield
    finally:
        kill_all_with_retries()
        ensure_free_disk_space()


def _human_readable_json_logs(raw_logs: str) -> str:
    output = []
    for line in raw_logs.splitlines():
        try:
            log_entry = json.loads(line)
            human_readable_log = " ".join(
                f"{key}: {value}" for key, value in log_entry.items()
            )
            output.append(f"\t{human_readable_log}")
        except Exception:
            output.append(line)
    return "\n".join(output)


@contextmanager
def _show_container_logs_if_raised():
    initial_ids = {c.id for c in get_containers({TRUSS: True})}
    exception_raised = False
    try:
        yield
    except Exception:
        exception_raised = True
        raise
    finally:
        if exception_raised:
            print("An exception was raised, showing logs of all containers.")
            containers = get_containers({TRUSS: True})
            new_containers = [c for c in containers if c.id not in initial_ids]
            for container in new_containers:
                print(f"Logs for container {container.name} ({container.id}):")
                print(_human_readable_json_logs(container.logs()))


def kill_all_with_retries(num_retries: int = 10):
    kill_all()
    attempts = 0
    while attempts < num_retries:
        containers = get_containers({TRUSS: True})
        if len(containers) == 0:
            return
        attempts += 1
        time.sleep(1)


def ensure_free_disk_space():
    """Check if disk space is low."""
    if is_disk_space_low():
        clear_disk_space()


def is_disk_space_low() -> bool:
    disk_usage = shutil.disk_usage("/")
    disk_free_percent = 100 * float(disk_usage.free) / disk_usage.total
    return disk_free_percent <= DISK_SPACE_LOW_PERCENTAGE


def clear_disk_space():
    docker_ps_output = subprocess.check_output(
        ["docker", "ps", "-a", "-f", "status=exited", "-q"]
    ).decode("utf-8")
    docker_containers = docker_ps_output.split("\n")[:-1]
    subprocess.run(["docker", "rm", *docker_containers])
    subprocess.run(["docker", "system", "prune", "-a", "-f"])



import json
import logging
import warnings
from enum import Enum
from typing import Optional

from huggingface_hub.errors import HFValidationError
from huggingface_hub.utils import validate_repo_id
from pydantic import BaseModel, PydanticDeprecatedSince20, validator
from rich.console import Console

# Suppress Pydantic V1 warnings, because we have to use it for backwards compat.
warnings.filterwarnings("ignore", category=PydanticDeprecatedSince20)

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

console = Console()


class TrussTRTLLMModel(str, Enum):
    LLAMA = "llama"
    MISTRAL = "mistral"
    DEEPSEEK = "deepseek"
    WHISPER = "whisper"


class TrussTRTLLMQuantizationType(str, Enum):
    NO_QUANT = "no_quant"
    WEIGHTS_ONLY_INT8 = "weights_int8"
    WEIGHTS_KV_INT8 = "weights_kv_int8"
    WEIGHTS_ONLY_INT4 = "weights_int4"
    WEIGHTS_KV_INT4 = "weights_kv_int4"
    SMOOTH_QUANT = "smooth_quant"
    FP8 = "fp8"
    FP8_KV = "fp8_kv"


class TrussTRTLLMPluginConfiguration(BaseModel):
    paged_kv_cache: bool = True
    gemm_plugin: str = "auto"
    use_paged_context_fmha: bool = False
    use_fp8_context_fmha: bool = False


class CheckpointSource(str, Enum):
    HF: str = "HF"
    GCS: str = "GCS"
    LOCAL: str = "LOCAL"
    # REMOTE_URL is useful when the checkpoint lives on remote storage accessible via HTTP (e.g a presigned URL)
    REMOTE_URL: str = "REMOTE_URL"


class CheckpointRepository(BaseModel):
    source: CheckpointSource
    repo: str


class TrussTRTLLMBuildConfiguration(BaseModel):
    base_model: TrussTRTLLMModel
    max_input_len: int
    max_output_len: int
    max_batch_size: int
    max_beam_width: int = 1
    max_prompt_embedding_table_size: int = 0
    checkpoint_repository: CheckpointRepository
    gather_all_token_logits: bool = False
    strongly_typed: bool = False
    quantization_type: TrussTRTLLMQuantizationType = (
        TrussTRTLLMQuantizationType.NO_QUANT
    )
    tensor_parallel_count: int = 1
    pipeline_parallel_count: int = 1
    plugin_configuration: TrussTRTLLMPluginConfiguration = (
        TrussTRTLLMPluginConfiguration()
    )
    use_fused_mlp: bool = False
    kv_cache_free_gpu_mem_fraction: float = 0.9
    num_builder_gpus: Optional[int] = None

    @validator("max_beam_width")
    def check_max_beam_width(cls, v: int):
        if isinstance(v, int):
            if v != 1:
                raise ValueError(
                    "max_beam_width greater than 1 is not currently supported"
                )
        return v


class TrussTRTLLMServingConfiguration(BaseModel):
    engine_repository: str
    tokenizer_repository: str
    tensor_parallel_count: int = 1
    pipeline_parallel_count: int = 1


class TRTLLMConfiguration(BaseModel):
    serve: Optional[TrussTRTLLMServingConfiguration] = None
    build: Optional[TrussTRTLLMBuildConfiguration] = None

    def __init__(self, **data):
        super().__init__(**data)
        self._validate_minimum_required_configuration()
        self._validate_kv_cache_flags()
        if self.build.checkpoint_repository.source == CheckpointSource.HF:
            self._validate_hf_repo_id()

    # In pydantic v2 this would be `@model_validator(mode="after")` and
    # the __init__ override can be removed.
    def _validate_minimum_required_configuration(self):
        if not self.serve and not self.build:
            raise ValueError("Either serve or build configurations must be provided")
        if self.serve and self.build:
            raise ValueError("Both serve and build configurations cannot be provided")
        if self.serve is not None:
            if (self.serve.engine_repository is None) ^ (
                self.serve.tokenizer_repository is None
            ):
                raise ValueError(
                    "Both engine_repository and tokenizer_repository must be provided"
                )
        return self

    def _validate_kv_cache_flags(self):
        if self.build is None:
            return self
        if not self.build.plugin_configuration.paged_kv_cache and (
            self.build.plugin_configuration.use_paged_context_fmha
            or self.build.plugin_configuration.use_fp8_context_fmha
        ):
            raise ValueError(
                "Using paged context fmha or fp8 context fmha requires requires paged kv cache"
            )
        if (
            self.build.plugin_configuration.use_fp8_context_fmha
            and not self.build.plugin_configuration.use_paged_context_fmha
        ):
            raise ValueError("Using fp8 context fmha requires paged context fmha")
        return self

    def _validate_hf_repo_id(self):
        try:
            validate_repo_id(self.build.checkpoint_repository.repo)
        except HFValidationError as e:
            raise ValueError(
                f"HuggingFace repository validation failed: {str(e)}"
            ) from e

    @property
    def requires_build(self):
        if self.build is not None:
            return True
        return False

    # TODO(Abu): Replace this with model_dump(json=True)
    # when pydantic v2 is used here
    def to_json_dict(self, verbose=True):
        return json.loads(self.json(exclude_unset=not verbose))



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Type

import boto3
import yaml
from botocore import UNSIGNED
from botocore.client import Config
from google.cloud import storage
from huggingface_hub import get_hf_file_metadata, hf_hub_url, list_repo_files
from huggingface_hub.utils import filter_repo_objects
from truss import constants
from truss.config.trt_llm import TrussTRTLLMModel
from truss.constants import (
    AUDIO_MODEL_TRTLLM_REQUIREMENTS,
    AUDIO_MODEL_TRTLLM_SYSTEM_PACKAGES,
    AUDIO_MODEL_TRTLLM_TRUSS_DIR,
    BASE_SERVER_REQUIREMENTS_TXT_FILENAME,
    BASE_TRTLLM_REQUIREMENTS,
    CONTROL_SERVER_CODE_DIR,
    DOCKER_SERVER_TEMPLATES_DIR,
    FILENAME_CONSTANTS_MAP,
    MAX_SUPPORTED_PYTHON_VERSION_IN_CUSTOM_BASE_IMAGE,
    MIN_SUPPORTED_PYTHON_VERSION_IN_CUSTOM_BASE_IMAGE,
    MODEL_DOCKERFILE_NAME,
    OPENAI_COMPATIBLE_TAG,
    REQUIREMENTS_TXT_FILENAME,
    SERVER_CODE_DIR,
    SERVER_DOCKERFILE_TEMPLATE_NAME,
    SERVER_REQUIREMENTS_TXT_FILENAME,
    SHARED_SERVING_AND_TRAINING_CODE_DIR,
    SHARED_SERVING_AND_TRAINING_CODE_DIR_NAME,
    SYSTEM_PACKAGES_TXT_FILENAME,
    TEMPLATES_DIR,
    TRTLLM_BASE_IMAGE,
    TRTLLM_PREDICT_CONCURRENCY,
    TRTLLM_PYTHON_EXECUTABLE,
    TRTLLM_TRUSS_DIR,
    USER_SUPPLIED_REQUIREMENTS_TXT_FILENAME,
)
from truss.contexts.image_builder.cache_warmer import (
    AWSCredentials,
    parse_s3_credentials_file,
)
from truss.contexts.image_builder.image_builder import ImageBuilder
from truss.contexts.image_builder.util import (
    TRUSS_BASE_IMAGE_VERSION_TAG,
    file_is_not_empty,
    to_dotted_python_version,
    truss_base_image_name,
    truss_base_image_tag,
)
from truss.contexts.truss_context import TrussContext
from truss.patch.hash import directory_content_hash
from truss.truss_config import DEFAULT_BUNDLED_PACKAGES_DIR, BaseImage, TrussConfig
from truss.truss_spec import TrussSpec
from truss.util.jinja import read_template_from_fs
from truss.util.path import (
    build_truss_target_directory,
    copy_tree_or_file,
    copy_tree_path,
    load_trussignore_patterns,
)

BUILD_SERVER_DIR_NAME = "server"
BUILD_CONTROL_SERVER_DIR_NAME = "control"
BUILD_SERVER_EXTENSIONS_PATH = "extensions"

CONFIG_FILE = "config.yaml"
USER_TRUSS_IGNORE_FILE = ".truss_ignore"
GCS_CREDENTIALS = "service_account.json"
S3_CREDENTIALS = "s3_credentials.json"

HF_ACCESS_TOKEN_FILE_NAME = "hf-access-token"

CLOUD_BUCKET_CACHE = Path("/app/model_cache/")
HF_SOURCE_DIR = Path("./root/.cache/huggingface/hub/")
HF_CACHE_DIR = Path("/root/.cache/huggingface/hub/")


class RemoteCache(ABC):
    def __init__(self, repo_name, data_dir, revision=None):
        self.repo_name = repo_name
        self.data_dir = data_dir
        self.revision = revision

    @staticmethod
    def from_repo(repo_name: str, data_dir: Path) -> "RemoteCache":
        repository_class: Type["RemoteCache"]
        if repo_name.startswith("gs://"):
            repository_class = GCSCache
        elif repo_name.startswith("s3://"):
            repository_class = S3Cache
        else:
            repository_class = HuggingFaceCache
        return repository_class(repo_name, data_dir)

    def filter(self, allow_patterns, ignore_patterns):
        return list(
            filter_repo_objects(
                items=self.list_files(revision=self.revision),
                allow_patterns=allow_patterns,
                ignore_patterns=ignore_patterns,
            )
        )

    @abstractmethod
    def list_files(self, revision=None):
        pass

    @abstractmethod
    def prepare_for_cache(self, filenames):
        pass


class GCSCache(RemoteCache):
    def list_files(self, revision=None):
        gcs_credentials_file = self.data_dir / GCS_CREDENTIALS

        if gcs_credentials_file.exists():
            storage_client = storage.Client.from_service_account_json(
                gcs_credentials_file
            )
        else:
            storage_client = storage.Client.create_anonymous_client()

        bucket_name, prefix = split_path(self.repo_name, prefix="gs://")
        blobs = storage_client.list_blobs(bucket_name, prefix=prefix)

        all_objects = []
        for blob in blobs:
            # leave out folders
            if blob.name[-1] == "/":
                continue
            all_objects.append(blob.name)

        return all_objects

    def prepare_for_cache(self, filenames):
        bucket_name, _ = split_path(self.repo_name, prefix="gs://")

        files_to_cache = []
        for filename in filenames:
            file_location = str(CLOUD_BUCKET_CACHE / bucket_name / filename)
            files_to_cache.append(CachedFile(source=file_location, dst=file_location))

        return files_to_cache


class S3Cache(RemoteCache):
    def list_files(self, revision=None):
        s3_credentials_file = self.data_dir / S3_CREDENTIALS

        if s3_credentials_file.exists():
            s3_credentials: AWSCredentials = parse_s3_credentials_file(
                self.data_dir / S3_CREDENTIALS
            )
            session = boto3.Session(
                aws_access_key_id=s3_credentials.access_key_id,
                aws_secret_access_key=s3_credentials.secret_access_key,
                aws_session_token=s3_credentials.session_token,
                region_name=s3_credentials.region,
            )
            s3 = session.resource("s3")
        else:
            s3 = boto3.resource("s3", config=Config(signature_version=UNSIGNED))

        # path may be like "folderA/folderB"
        bucket_name, path = split_path(self.repo_name, prefix="s3://")
        bucket = s3.Bucket(bucket_name)

        all_objects = []
        for blob in bucket.objects.filter(Prefix=path):
            # leave out folders
            if blob.key[-1] == "/":
                continue
            all_objects.append(blob.key)

        return all_objects

    def prepare_for_cache(self, filenames):
        bucket_name, _ = split_path(self.repo_name, prefix="s3://")

        files_to_cache = []
        for filename in filenames:
            file_location = str(CLOUD_BUCKET_CACHE / bucket_name / filename)
            files_to_cache.append(CachedFile(source=file_location, dst=file_location))

        return files_to_cache


def hf_cache_file_from_location(path: str):
    src_file_location = str(HF_SOURCE_DIR / path)
    dst_file_location = str(HF_CACHE_DIR / path)
    cache_file = CachedFile(source=src_file_location, dst=dst_file_location)
    return cache_file


class HuggingFaceCache(RemoteCache):
    def list_files(self, revision=None):
        return list_repo_files(self.repo_name, revision=revision)

    def prepare_for_cache(self, filenames):
        files_to_cache = []
        repo_folder_name = f"models--{self.repo_name.replace('/', '--')}"
        for filename in filenames:
            hf_url = hf_hub_url(self.repo_name, filename)
            hf_file_metadata = get_hf_file_metadata(hf_url)

            files_to_cache.append(
                hf_cache_file_from_location(
                    f"{repo_folder_name}/blobs/{hf_file_metadata.etag}"
                )
            )

        # snapshots is just a set of folders with symlinks -- we can copy the entire thing separately
        files_to_cache.append(
            hf_cache_file_from_location(f"{repo_folder_name}/snapshots/")
        )

        # refs just has files with revision commit hashes
        files_to_cache.append(hf_cache_file_from_location(f"{repo_folder_name}/refs/"))

        files_to_cache.append(hf_cache_file_from_location("version.txt"))

        return files_to_cache


def get_credentials_to_cache(data_dir: Path) -> List[str]:
    gcs_credentials_file = data_dir / GCS_CREDENTIALS
    s3_credentials_file = data_dir / S3_CREDENTIALS
    credentials = [gcs_credentials_file, s3_credentials_file]

    credentials_to_cache = []
    for file in credentials:
        if file.exists():
            build_path = Path(*file.parts[-2:])
            credentials_to_cache.append(str(build_path))

    return credentials_to_cache


def split_path(path, prefix="gs://"):
    # Remove the 'gs://' prefix
    path = path.replace(prefix, "")

    # Split on the first slash
    parts = path.split("/", 1)

    bucket_name = parts[0]
    path = parts[1] if len(parts) > 1 else ""

    return bucket_name, path


@dataclass
class CachedFile:
    source: str
    dst: str


def get_files_to_cache(config: TrussConfig, truss_dir: Path, build_dir: Path):
    def copy_into_build_dir(from_path: Path, path_in_build_dir: str):
        copy_tree_or_file(from_path, build_dir / path_in_build_dir)  # type: ignore[operator]

    remote_model_files = {}
    local_files_to_cache: List[CachedFile] = []
    if config.model_cache:
        curr_dir = Path(__file__).parent.resolve()
        copy_into_build_dir(curr_dir / "cache_warmer.py", "cache_warmer.py")
        for model in config.model_cache.models:
            repo_id = model.repo_id
            revision = model.revision

            allow_patterns = model.allow_patterns
            ignore_patterns = model.ignore_patterns

            model_cache = RemoteCache.from_repo(repo_id, truss_dir / config.data_dir)
            remote_filtered_files = model_cache.filter(allow_patterns, ignore_patterns)
            local_files_to_cache += model_cache.prepare_for_cache(remote_filtered_files)

            remote_model_files[repo_id] = {
                "files": remote_filtered_files,
                "revision": revision,
            }

    copy_into_build_dir(
        TEMPLATES_DIR / "cache_requirements.txt", "cache_requirements.txt"
    )
    return remote_model_files, local_files_to_cache


def update_config_and_gather_files(
    config: TrussConfig, truss_dir: Path, build_dir: Path
):
    return get_files_to_cache(config, truss_dir, build_dir)


def generate_docker_server_nginx_config(build_dir, config):
    nginx_template = read_template_from_fs(
        DOCKER_SERVER_TEMPLATES_DIR, "proxy.conf.jinja"
    )

    nginx_content = nginx_template.render(
        server_endpoint=config.docker_server.predict_endpoint,
        readiness_endpoint=config.docker_server.readiness_endpoint,
        liveness_endpoint=config.docker_server.liveness_endpoint,
        server_port=config.docker_server.server_port,
    )
    nginx_filepath = build_dir / "proxy.conf"
    nginx_filepath.write_text(nginx_content)


def generate_docker_server_supervisord_config(build_dir, config):
    supervisord_template = read_template_from_fs(
        DOCKER_SERVER_TEMPLATES_DIR, "supervisord.conf.jinja"
    )
    supervisord_contents = supervisord_template.render(
        start_command=config.docker_server.start_command,
        setup_command=config.docker_server.setup_command,
    )
    supervisord_filepath = build_dir / "supervisord.conf"
    supervisord_filepath.write_text(supervisord_contents)


class ServingImageBuilderContext(TrussContext):
    @staticmethod
    def run(truss_dir: Path):
        return ServingImageBuilder(truss_dir)


class ServingImageBuilder(ImageBuilder):
    def __init__(self, truss_dir: Path) -> None:
        self._truss_dir = truss_dir
        self._spec = TrussSpec(truss_dir)

    @property
    def default_tag(self):
        return f"{self._spec.model_framework_name}-model:latest"

    def prepare_image_build_dir(
        self, build_dir: Optional[Path] = None, use_hf_secret: bool = False
    ):
        """
        Prepare a directory for building the docker image from.
        """
        truss_dir = self._truss_dir
        spec = self._spec
        config = spec.config
        model_framework_name = spec.model_framework_name
        if build_dir is None:
            # TODO(pankaj) We probably don't need model framework specific directory.
            build_dir = build_truss_target_directory(model_framework_name)

        data_dir = build_dir / config.data_dir  # type: ignore[operator]

        def copy_into_build_dir(from_path: Path, path_in_build_dir: str):
            copy_tree_or_file(from_path, build_dir / path_in_build_dir)  # type: ignore[operator]

        truss_ignore_patterns = []
        if (truss_dir / USER_TRUSS_IGNORE_FILE).exists():
            truss_ignore_patterns = load_trussignore_patterns(
                truss_dir / USER_TRUSS_IGNORE_FILE
            )

        # Copy over truss
        copy_tree_path(truss_dir, build_dir, ignore_patterns=truss_ignore_patterns)

        if config.docker_server is not None:
            copy_into_build_dir(
                TEMPLATES_DIR / "docker_server_requirements.txt",
                "docker_server_requirements.txt",
            )

            generate_docker_server_nginx_config(build_dir, config)

            generate_docker_server_supervisord_config(build_dir, config)

        # Copy over template truss for TRT-LLM (we overwrite the model and packages dir)
        # Most of the code is pulled from upstream triton-inference-server tensorrtllm_backend
        # https://github.com/triton-inference-server/tensorrtllm_backend/tree/v0.9.0/all_models/inflight_batcher_llm
        if config.trt_llm is not None:
            is_audio_model = (
                config.trt_llm.build.base_model == TrussTRTLLMModel.WHISPER
                if config.trt_llm.build is not None
                else False
            )

            if is_audio_model:
                copy_tree_path(
                    AUDIO_MODEL_TRTLLM_TRUSS_DIR, build_dir, ignore_patterns=[]
                )
            else:
                # trt_llm is treated as an extension at model run time.
                copy_into_build_dir(
                    TRTLLM_TRUSS_DIR / "src",
                    f"{BUILD_SERVER_DIR_NAME}/{BUILD_SERVER_EXTENSIONS_PATH}/trt_llm",
                )
                # TODO(pankaj) Do this differently. This is not ideal, user
                # supplied code in bundled packages can conflict with those from
                # the trtllm extension. We don't want to put this in the build
                # directory directly either because of chances of conflict there
                # as well and the noise it can create there. We need to find a
                # new place that's made available in model's pythonpath. This is
                # a bigger lift and feels overkill right now. Worth revisiting
                # if we come across cases of actual conflicts.
                copy_into_build_dir(
                    TRTLLM_TRUSS_DIR / DEFAULT_BUNDLED_PACKAGES_DIR,
                    DEFAULT_BUNDLED_PACKAGES_DIR,
                )

            tensor_parallel_count = (
                config.trt_llm.build.tensor_parallel_count  # type: ignore[union-attr]
                if config.trt_llm.build is not None
                else config.trt_llm.serve.tensor_parallel_count  # type: ignore[union-attr]
            )

            if tensor_parallel_count != config.resources.accelerator.count:
                raise ValueError(
                    "Tensor parallelism and GPU count must be the same for TRT-LLM"
                )

            config.runtime.predict_concurrency = TRTLLM_PREDICT_CONCURRENCY

            if not is_audio_model:
                config.base_image = BaseImage(
                    image=TRTLLM_BASE_IMAGE,
                    python_executable_path=TRTLLM_PYTHON_EXECUTABLE,
                )

                config.requirements.extend(BASE_TRTLLM_REQUIREMENTS)

                config.model_metadata["tags"] = [OPENAI_COMPATIBLE_TAG]
            else:
                config.requirements.extend(AUDIO_MODEL_TRTLLM_REQUIREMENTS)
                config.system_packages.extend(AUDIO_MODEL_TRTLLM_SYSTEM_PACKAGES)
                config.python_version = "py310"

        # Override config.yml
        with (build_dir / CONFIG_FILE).open("w") as config_file:
            yaml.dump(config.to_dict(verbose=True), config_file)

        external_data_files: list = []
        data_dir = Path("/app/data/")
        if self._spec.external_data is not None:
            for ext_file in self._spec.external_data.items:
                external_data_files.append(
                    (
                        ext_file.url,
                        (data_dir / ext_file.local_data_path).resolve(),
                    )
                )

        # Download from HuggingFace
        model_files, cached_files = update_config_and_gather_files(
            config, truss_dir, build_dir
        )

        # Copy inference server code
        copy_into_build_dir(SERVER_CODE_DIR, BUILD_SERVER_DIR_NAME)
        copy_into_build_dir(
            SHARED_SERVING_AND_TRAINING_CODE_DIR,
            BUILD_SERVER_DIR_NAME + "/" + SHARED_SERVING_AND_TRAINING_CODE_DIR_NAME,
        )

        # Copy control server code
        if config.live_reload:
            copy_into_build_dir(CONTROL_SERVER_CODE_DIR, BUILD_CONTROL_SERVER_DIR_NAME)
            copy_into_build_dir(
                SHARED_SERVING_AND_TRAINING_CODE_DIR,
                BUILD_CONTROL_SERVER_DIR_NAME
                + "/control/"
                + SHARED_SERVING_AND_TRAINING_CODE_DIR_NAME,
            )

        # Copy base TrussServer requirements if supplied custom base image
        base_truss_server_reqs_filepath = SERVER_CODE_DIR / REQUIREMENTS_TXT_FILENAME
        if config.base_image:
            copy_into_build_dir(
                base_truss_server_reqs_filepath, BASE_SERVER_REQUIREMENTS_TXT_FILENAME
            )

        # Copy model framework specific requirements file
        server_reqs_filepath = (
            TEMPLATES_DIR / model_framework_name / REQUIREMENTS_TXT_FILENAME
        )
        should_install_server_requirements = file_is_not_empty(server_reqs_filepath)
        if should_install_server_requirements:
            copy_into_build_dir(server_reqs_filepath, SERVER_REQUIREMENTS_TXT_FILENAME)

        with open(base_truss_server_reqs_filepath, "r") as f:
            base_server_requirements = f.read()

        if config.docker_server:
            # when docker server is enabled, no need to install truss requirements
            # only install user-provided python requirements
            user_provided_python_requirements = spec.requirements_txt
        else:
            # If the user has provided python requirements,
            # append the truss server requirements, so that any conflicts
            # are detected and cause a build failure. If there are no
            # requirements provided, we just pass an empty string,
            # as there's no need to install anything.
            # TODO (BT-10217): above reasoning leads to inconsistencies. To get consistent
            #  images tentatively add server requirements always. This whole point needs
            #  more thought and potentially a re-design.
            user_provided_python_requirements = (
                base_server_requirements + spec.requirements_txt
                if spec.requirements
                else base_server_requirements
            )
        if spec.requirements_file is not None:
            copy_into_build_dir(
                truss_dir / spec.requirements_file,
                USER_SUPPLIED_REQUIREMENTS_TXT_FILENAME,
            )
        (build_dir / REQUIREMENTS_TXT_FILENAME).write_text(
            user_provided_python_requirements
        )
        (build_dir / SYSTEM_PACKAGES_TXT_FILENAME).write_text(spec.system_packages_txt)

        self._render_dockerfile(
            build_dir,
            should_install_server_requirements,
            model_files,
            use_hf_secret,
            cached_files,
            external_data_files,
            self._spec.build_commands,
        )

    def _render_dockerfile(
        self,
        build_dir: Path,
        should_install_server_requirements: bool,
        model_files: Dict[str, Any],
        use_hf_secret: bool,
        cached_files: List[str],
        external_data_files: List[Tuple[str, str]],
        build_commands: List[str],
    ):
        config = self._spec.config
        data_dir = build_dir / config.data_dir
        model_dir = build_dir / config.model_module_dir
        bundled_packages_dir = build_dir / config.bundled_packages_dir
        dockerfile_template = read_template_from_fs(
            TEMPLATES_DIR, SERVER_DOCKERFILE_TEMPLATE_NAME
        )
        python_version = to_dotted_python_version(config.python_version)
        if config.base_image:
            base_image_name_and_tag = config.base_image.image
        else:
            base_image_name = truss_base_image_name(job_type="server")
            tag = truss_base_image_tag(
                python_version=python_version,
                use_gpu=config.resources.use_gpu,
                version_tag=TRUSS_BASE_IMAGE_VERSION_TAG,
            )
            base_image_name_and_tag = f"{base_image_name}:{tag}"
        should_install_system_requirements = file_is_not_empty(
            build_dir / SYSTEM_PACKAGES_TXT_FILENAME
        )
        should_install_python_requirements = file_is_not_empty(
            build_dir / REQUIREMENTS_TXT_FILENAME
        )
        should_install_user_requirements_file = file_is_not_empty(
            build_dir / USER_SUPPLIED_REQUIREMENTS_TXT_FILENAME
        )

        max_supported_python_version_in_custom_base_image = (
            MAX_SUPPORTED_PYTHON_VERSION_IN_CUSTOM_BASE_IMAGE
        )
        min_supported_python_version_in_custom_base_image = (
            MIN_SUPPORTED_PYTHON_VERSION_IN_CUSTOM_BASE_IMAGE
        )
        max_supported_python_minor_version_in_custom_base_image = (
            MAX_SUPPORTED_PYTHON_VERSION_IN_CUSTOM_BASE_IMAGE.split(".")[1]
        )
        min_supported_python_minor_version_in_custom_base_image = (
            MIN_SUPPORTED_PYTHON_VERSION_IN_CUSTOM_BASE_IMAGE.split(".")[1]
        )
        supported_python_major_version_in_custom_base_image = (
            MIN_SUPPORTED_PYTHON_VERSION_IN_CUSTOM_BASE_IMAGE.split(".")[0]
        )

        hf_access_token = config.secrets.get(constants.HF_ACCESS_TOKEN_KEY)
        dockerfile_contents = dockerfile_template.render(
            should_install_server_requirements=should_install_server_requirements,
            base_image_name_and_tag=base_image_name_and_tag,
            max_supported_python_version_in_custom_base_image=max_supported_python_version_in_custom_base_image,
            min_supported_python_version_in_custom_base_image=min_supported_python_version_in_custom_base_image,
            max_supported_python_minor_version_in_custom_base_image=max_supported_python_minor_version_in_custom_base_image,
            min_supported_python_minor_version_in_custom_base_image=min_supported_python_minor_version_in_custom_base_image,
            supported_python_major_version_in_custom_base_image=supported_python_major_version_in_custom_base_image,
            should_install_system_requirements=should_install_system_requirements,
            should_install_requirements=should_install_python_requirements,
            should_install_user_requirements_file=should_install_user_requirements_file,
            config=config,
            python_version=python_version,
            live_reload=config.live_reload,
            data_dir_exists=data_dir.exists(),
            model_dir_exists=model_dir.exists(),
            bundled_packages_dir_exists=bundled_packages_dir.exists(),
            truss_hash=directory_content_hash(
                self._truss_dir, self._spec.hash_ignore_patterns
            ),
            models=model_files,
            use_hf_secret=use_hf_secret,
            cached_files=cached_files,
            credentials_to_cache=get_credentials_to_cache(data_dir),
            model_cache=len(config.model_cache.models) > 0,
            hf_access_token=hf_access_token,
            hf_access_token_file_name=HF_ACCESS_TOKEN_FILE_NAME,
            external_data_files=external_data_files,
            build_commands=build_commands,
            **FILENAME_CONSTANTS_MAP,
        )
        docker_file_path = build_dir / MODEL_DOCKERFILE_NAME
        docker_file_path.write_text(dockerfile_contents)



import random

# For more on chains, check out https://docs.baseten.co/chains/overview.
import truss_chains as chains


# By inhereting chains.ChainletBase, the chains framework will know to create a chainlet that hosts the RandInt class.
class RandInt(chains.ChainletBase):
    # run_remote must be implemented by all chainlets. This is the code that will be executed at inference time.
    def run_remote(self, max_value: int) -> int:
        return random.randint(1, max_value)


# The @chains.mark_entrypoint decorator indicates that this Chainlet is the entrypoint.
# Each chain must have exactly one entrypoint.
@chains.mark_entrypoint
class HelloWorld(chains.ChainletBase):
    # chains.depends indicates that the HelloWorld chainlet depends on the RandInt Chainlet
    # this enables the HelloWorld chainlet to call the RandInt chainlet
    def __init__(self, rand_int=chains.depends(RandInt, retries=3)) -> None:
        self._rand_int = rand_int

    def run_remote(self, max_value: int) -> str:
        num_repetitions = self._rand_int.run_remote(max_value)
        return "Hello World! " * num_repetitions


if __name__ == "__main__":
    with chains.run_local():
        hello_world_chain = HelloWorld()
        result = hello_world_chain.run_remote(max_value=5)

    print(result)
    # Hello World! Hello World! Hello World!



# flake8: noqa
# type: ignore

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: briton.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0c\x62riton.proto\x12\x06\x62riton"r\n\x06Tensor\x12#\n\x05shape\x18\x01 \x01(\x0b\x32\x14.briton.Tensor.Shape\x12\x1f\n\x05\x64type\x18\x02 \x01(\x0e\x32\x10.briton.DataType\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\x0c\x1a\x14\n\x05Shape\x12\x0b\n\x03\x64im\x18\x01 \x03(\x03"\x82\x07\n\x10InferenceRequest\x12\x12\n\nrequest_id\x18\x01 \x01(\x03\x12\x12\n\ninput_text\x18\x02 \x01(\t\x12\x11\n\tinput_ids\x18\x03 \x03(\x05\x12\x1f\n\x12request_output_len\x18\x05 \x01(\rH\x00\x88\x01\x01\x12\x13\n\x06\x65nd_id\x18\x06 \x01(\rH\x01\x88\x01\x01\x12\x13\n\x06pad_id\x18\x07 \x01(\rH\x02\x88\x01\x01\x12\x17\n\nbeam_width\x18\n \x01(\rH\x03\x88\x01\x01\x12\x18\n\x0btemperature\x18\x0b \x01(\x02H\x04\x88\x01\x01\x12\x1a\n\rruntime_top_k\x18\x0c \x01(\rH\x05\x88\x01\x01\x12\x1a\n\rruntime_top_p\x18\r \x01(\x02H\x06\x88\x01\x01\x12\x18\n\x0blen_penalty\x18\x0e \x01(\x02H\x07\x88\x01\x01\x12\x1f\n\x12repetition_penalty\x18\x0f \x01(\x02H\x08\x88\x01\x01\x12\x1d\n\x10presence_penalty\x18\x10 \x01(\x02H\t\x88\x01\x01\x12\x11\n\tbad_words\x18\x11 \x03(\t\x12\x12\n\nstop_words\x18\x12 \x03(\t\x12\x19\n\x0clora_task_id\x18\x13 \x01(\x04H\n\x88\x01\x01\x12)\n\x0clora_weights\x18\x14 \x01(\x0b\x32\x0e.briton.TensorH\x0b\x88\x01\x01\x12(\n\x0blora_config\x18\x15 \x01(\x0b\x32\x0e.briton.TensorH\x0c\x88\x01\x01\x12\x18\n\x0brandom_seed\x18\x16 \x01(\x03H\r\x88\x01\x01\x12\x1f\n\x12output_schema_hash\x18\x17 \x01(\tH\x0e\x88\x01\x01\x12\x15\n\x08tools_id\x18\x18 \x01(\rH\x0f\x88\x01\x01\x12\x18\n\x0b\x66orce_tools\x18\x19 \x01(\x08H\x10\x88\x01\x01\x42\x15\n\x13_request_output_lenB\t\n\x07_end_idB\t\n\x07_pad_idB\r\n\x0b_beam_widthB\x0e\n\x0c_temperatureB\x10\n\x0e_runtime_top_kB\x10\n\x0e_runtime_top_pB\x0e\n\x0c_len_penaltyB\x15\n\x13_repetition_penaltyB\x13\n\x11_presence_penaltyB\x0f\n\r_lora_task_idB\x0f\n\r_lora_weightsB\x0e\n\x0c_lora_configB\x0e\n\x0c_random_seedB\x15\n\x13_output_schema_hashB\x0b\n\t_tools_idB\x0e\n\x0c_force_tools"R\n\x13InferenceAnswerPart\x12\x12\n\nrequest_id\x18\x01 \x01(\x03\x12\x13\n\x0boutput_text\x18\x02 \x01(\t\x12\x12\n\noutput_ids\x18\x03 \x03(\x05"\xa6\x08\n\x0c\x42ritonConfig\x12\x13\n\x0b\x65ngine_path\x18\x01 \x01(\t\x12\x14\n\x0chf_tokenizer\x18\x02 \x01(\t\x12N\n\x16\x62\x61tch_scheduler_policy\x18\x05 \x01(\x0e\x32).briton.BritonConfig.BatchSchedulerPolicyH\x00\x88\x01\x01\x12\x1f\n\x12\x65nable_trt_overlap\x18\x06 \x01(\x08H\x01\x88\x01\x01\x12)\n\x1cmax_tokens_in_paged_kv_cache\x18\n \x01(\x04H\x02\x88\x01\x01\x12+\n\x1ekv_cache_free_gpu_mem_fraction\x18\x0b \x01(\x02H\x03\x88\x01\x01\x12!\n\x14medusa_decoding_mode\x18\x0c \x01(\x08H\x04\x88\x01\x01\x12#\n\x16\x65nable_chunked_context\x18\r \x01(\x08H\x05\x88\x01\x01\x12"\n\x15\x65nable_kv_cache_reuse\x18\x0e \x01(\x08H\x06\x88\x01\x01\x12\'\n\x1akv_cache_host_memory_bytes\x18\x0f \x01(\x04H\x07\x88\x01\x01\x12(\n\x1blora_cache_max_adapter_size\x18\x10 \x01(\x04H\x08\x88\x01\x01\x12,\n\x1flora_cache_optimal_adapter_size\x18\x11 \x01(\x04H\t\x88\x01\x01\x12+\n\x1elora_cache_gpu_memory_fraction\x18\x12 \x01(\x02H\n\x88\x01\x01\x12)\n\x1clora_cache_host_memory_bytes\x18\x13 \x01(\x04H\x0b\x88\x01\x01\x12\x1a\n\rfsm_cache_dir\x18\x14 \x01(\tH\x0c\x88\x01\x01"D\n\x14\x42\x61tchSchedulerPolicy\x12\x13\n\x0fMAX_UTILIZATION\x10\x00\x12\x17\n\x13GUARANTEED_NO_EVICT\x10\x01\x42\x19\n\x17_batch_scheduler_policyB\x15\n\x13_enable_trt_overlapB\x1f\n\x1d_max_tokens_in_paged_kv_cacheB!\n\x1f_kv_cache_free_gpu_mem_fractionB\x17\n\x15_medusa_decoding_modeB\x19\n\x17_enable_chunked_contextB\x18\n\x16_enable_kv_cache_reuseB\x1d\n\x1b_kv_cache_host_memory_bytesB\x1e\n\x1c_lora_cache_max_adapter_sizeB"\n _lora_cache_optimal_adapter_sizeB!\n\x1f_lora_cache_gpu_memory_fractionB\x1f\n\x1d_lora_cache_host_memory_bytesB\x10\n\x0e_fsm_cache_dir"\x98\x01\n\x10TokenToNextState\x12K\n\x13token_to_next_state\x18\x01 \x03(\x0b\x32..briton.TokenToNextState.TokenToNextStateEntry\x1a\x37\n\x15TokenToNextStateEntry\x12\x0b\n\x03key\x18\x01 \x01(\x05\x12\r\n\x05value\x18\x02 \x01(\x05:\x02\x38\x01"\xfb\x01\n\x0eStatesToTokens\x12\x44\n\x10states_to_tokens\x18\x01 \x03(\x0b\x32*.briton.StatesToTokens.StatesToTokensEntry\x12\x17\n\nvocab_size\x18\x02 \x01(\x05H\x00\x88\x01\x01\x12\x19\n\x0c\x65os_token_id\x18\x03 \x01(\x05H\x01\x88\x01\x01\x1aO\n\x13StatesToTokensEntry\x12\x0b\n\x03key\x18\x01 \x01(\x05\x12\'\n\x05value\x18\x02 \x01(\x0b\x32\x18.briton.TokenToNextState:\x02\x38\x01\x42\r\n\x0b_vocab_sizeB\x0f\n\r_eos_token_id*\xa8\x01\n\x08\x44\x61taType\x12\x0e\n\nDT_INVALID\x10\x00\x12\x0b\n\x07\x44T_INT4\x10\x01\x12\x0b\n\x07\x44T_INT8\x10\x02\x12\x0c\n\x08\x44T_UINT8\x10\x03\x12\x0c\n\x08\x44T_INT32\x10\x04\x12\x0c\n\x08\x44T_INT64\x10\x05\x12\x0e\n\nDT_FLOAT16\x10\n\x12\x0f\n\x0b\x44T_BFLOAT16\x10\x0b\x12\x0e\n\nDT_FLOAT32\x10\x0c\x12\n\n\x06\x44T_FP8\x10\r\x12\x0b\n\x07\x44T_BOOL\x10\x14\x32L\n\x06\x42riton\x12\x42\n\x05Infer\x12\x18.briton.InferenceRequest\x1a\x1b.briton.InferenceAnswerPart"\x00\x30\x01\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "briton_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _globals["_TOKENTONEXTSTATE_TOKENTONEXTSTATEENTRY"]._options = None
    _globals["_TOKENTONEXTSTATE_TOKENTONEXTSTATEENTRY"]._serialized_options = b"8\001"
    _globals["_STATESTOTOKENS_STATESTOTOKENSENTRY"]._options = None
    _globals["_STATESTOTOKENS_STATESTOTOKENSENTRY"]._serialized_options = b"8\001"
    _globals["_DATATYPE"]._serialized_start = 2600
    _globals["_DATATYPE"]._serialized_end = 2768
    _globals["_TENSOR"]._serialized_start = 24
    _globals["_TENSOR"]._serialized_end = 138
    _globals["_TENSOR_SHAPE"]._serialized_start = 118
    _globals["_TENSOR_SHAPE"]._serialized_end = 138
    _globals["_INFERENCEREQUEST"]._serialized_start = 141
    _globals["_INFERENCEREQUEST"]._serialized_end = 1039
    _globals["_INFERENCEANSWERPART"]._serialized_start = 1041
    _globals["_INFERENCEANSWERPART"]._serialized_end = 1123
    _globals["_BRITONCONFIG"]._serialized_start = 1126
    _globals["_BRITONCONFIG"]._serialized_end = 2188
    _globals["_BRITONCONFIG_BATCHSCHEDULERPOLICY"]._serialized_start = 1739
    _globals["_BRITONCONFIG_BATCHSCHEDULERPOLICY"]._serialized_end = 1807
    _globals["_TOKENTONEXTSTATE"]._serialized_start = 2191
    _globals["_TOKENTONEXTSTATE"]._serialized_end = 2343
    _globals["_TOKENTONEXTSTATE_TOKENTONEXTSTATEENTRY"]._serialized_start = 2288
    _globals["_TOKENTONEXTSTATE_TOKENTONEXTSTATEENTRY"]._serialized_end = 2343
    _globals["_STATESTOTOKENS"]._serialized_start = 2346
    _globals["_STATESTOTOKENS"]._serialized_end = 2597
    _globals["_STATESTOTOKENS_STATESTOTOKENSENTRY"]._serialized_start = 2486
    _globals["_STATESTOTOKENS_STATESTOTOKENSENTRY"]._serialized_end = 2565
    _globals["_BRITON"]._serialized_start = 2770
    _globals["_BRITON"]._serialized_end = 2846
# @@protoc_insertion_point(module_scope)



import os
import signal
import socket
import sys
import tempfile
import time
from multiprocessing import Process
from pathlib import Path

import pytest


@pytest.mark.integration
def test_truss_server_termination(truss_container_fs):
    port = 10123

    def start_truss_server(stdout_capture_file_path):
        sys.stdout = open(stdout_capture_file_path, "w")
        app_path = truss_container_fs / "app"
        sys.path.append(str(app_path))

        from truss_server import TrussServer

        server = TrussServer(http_port=port, config_or_path=app_path / "config.yaml")
        server.start()

    stdout_capture_file = tempfile.NamedTemporaryFile()
    subproc = Process(target=start_truss_server, args=(stdout_capture_file.name,))
    subproc.start()
    proc_id = subproc.pid
    time.sleep(2.0)
    # Port should have been taken up by truss server
    assert not _is_port_available(port)
    os.kill(proc_id, signal.SIGTERM)
    time.sleep(2.0)
    # Print on purpose for help with debugging, otherwise hard to know what's going on
    print(Path(stdout_capture_file.name).read_text())
    assert not subproc.is_alive()
    # Port should be free now
    assert _is_port_available(port)


def _is_port_available(port):
    try:
        # Try to bind to the given port
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.bind(("localhost", port))
            return True
    except socket.error:
        # Port is already in use
        return False



import concurrent
import contextlib
import dataclasses
import inspect
import json
import logging
import pathlib
import tempfile
import textwrap
import time
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from threading import Thread
from typing import Iterator, Mapping

import httpx
import opentelemetry.trace.propagation.tracecontext as tracecontext
import pytest
import requests
from opentelemetry import context, trace
from requests.exceptions import RequestException

from truss.local.local_config_handler import LocalConfigHandler
from truss.model_inference import map_to_supported_python_version
from truss.tests.helpers import create_truss
from truss.tests.test_testing_utilities_for_other_tests import ensure_kill_all
from truss.truss_handle import TrussHandle

logger = logging.getLogger(__name__)

DEFAULT_LOG_ERROR = "Internal Server Error"
PREDICT_URL = "http://localhost:8090/v1/models/model:predict"


def _log_contains_error(line: dict, error: str, message: str):
    return (
        line["levelname"] == "ERROR"
        and line["message"] == message
        and error in line["exc_info"]
    )


def assert_logs_contain_error(logs: str, error: str, message=DEFAULT_LOG_ERROR):
    loglines = [json.loads(line) for line in logs.splitlines()]
    assert any(_log_contains_error(line, error, message) for line in loglines), (
        f"Did not find expected error in logs.\nExpected error: {error}\n"
        f"Expected message: {message}\nActual logs:\n{loglines}"
    )


class PropagatingThread(Thread):
    """
    PropagatingThread allows us to run threads and keep track of exceptions
    thrown.
    """

    def run(self):
        self.exc = None
        try:
            self.ret = self._target(*self._args, **self._kwargs)
        except BaseException as e:
            self.exc = e

    def join(self, timeout=None):
        super(PropagatingThread, self).join(timeout)
        if self.exc:
            raise self.exc
        return self.ret


@contextlib.contextmanager
def temp_truss(model_src: str, config_src: str) -> Iterator[TrussHandle]:
    with ensure_kill_all(), tempfile.TemporaryDirectory(dir=".") as tmp_work_dir:
        truss_dir = Path(tmp_work_dir, "truss")
        create_truss(truss_dir, config_src, textwrap.dedent(model_src))
        yield TrussHandle(truss_dir)


@pytest.mark.parametrize(
    "python_version, expected_python_version",
    [
        ("py37", "py38"),
        ("py38", "py38"),
        ("py39", "py39"),
        ("py310", "py310"),
        ("py311", "py311"),
        ("py312", "py311"),
        ("py36", "py38"),
    ],
)
def test_map_to_supported_python_version(python_version, expected_python_version):
    out_python_version = map_to_supported_python_version(python_version)
    assert out_python_version == expected_python_version


@pytest.mark.integration
def test_model_load_failure_truss():
    with ensure_kill_all():
        truss_root = Path(__file__).parent.parent.parent.resolve() / "truss"
        truss_dir = truss_root / "test_data" / "model_load_failure_test"
        tr = TrussHandle(truss_dir)

        _ = tr.docker_run(local_port=8090, detach=True, wait_for_server_ready=False)

        # Sleep a few seconds to get the server some time to  wake up
        time.sleep(10)

        truss_server_addr = "http://localhost:8090"

        def handle_request_exception(func):
            def wrapper(*args, **kwargs):
                try:
                    return func(*args, **kwargs)
                except RequestException:
                    return False

            return wrapper

        @handle_request_exception
        def _test_liveness_probe(expected_code):
            live = requests.get(f"{truss_server_addr}/")
            assert live.status_code == expected_code
            return True

        @handle_request_exception
        def _test_readiness_probe(expected_code):
            ready = requests.get(f"{truss_server_addr}/v1/models/model")
            assert ready.status_code == expected_code
            return True

        @handle_request_exception
        def _test_ping(expected_code):
            ping = requests.get(f"{truss_server_addr}/ping")
            assert ping.status_code == expected_code
            return True

        @handle_request_exception
        def _test_invocations(expected_code):
            invocations = requests.post(f"{truss_server_addr}/invocations", json={})
            assert invocations.status_code == expected_code
            return True

        # The server should be completely down so all requests should result in a RequestException.
        # The decorator handle_request_exception catches the RequestException and returns False.
        assert not _test_readiness_probe(expected_code=200)
        assert not _test_liveness_probe(expected_code=200)
        assert not _test_ping(expected_code=200)
        assert not _test_invocations(expected_code=200)


@pytest.mark.integration
def test_concurrency_truss():
    # Tests that concurrency limits work correctly
    with ensure_kill_all():
        truss_root = Path(__file__).parent.parent.parent.resolve() / "truss"
        truss_dir = truss_root / "test_data" / "test_concurrency_truss"
        tr = TrussHandle(truss_dir)
        _ = tr.docker_run(local_port=8090, detach=True, wait_for_server_ready=True)

        # Each request takes 2 seconds, for this thread, we allow
        # a concurrency of 2. This means the first two requests will
        # succeed within the 2 seconds, and the third will fail, since
        # it cannot start until the first two have completed.
        def make_request():
            requests.post(PREDICT_URL, json={}, timeout=3)

        successful_thread_1 = PropagatingThread(target=make_request)
        successful_thread_2 = PropagatingThread(target=make_request)
        failed_thread = PropagatingThread(target=make_request)

        successful_thread_1.start()
        successful_thread_2.start()
        # Ensure that the thread to fail starts a little after the others
        time.sleep(0.2)
        failed_thread.start()

        successful_thread_1.join()
        successful_thread_2.join()
        with pytest.raises(requests.exceptions.ReadTimeout):
            failed_thread.join()


@pytest.mark.integration
def test_requirements_file_truss():
    with ensure_kill_all():
        truss_root = Path(__file__).parent.parent.parent.resolve() / "truss"
        truss_dir = truss_root / "test_data" / "test_requirements_file_truss"
        tr = TrussHandle(truss_dir)
        _ = tr.docker_run(local_port=8090, detach=True, wait_for_server_ready=True)

        # The prediction imports torch which is specified in a requirements.txt and returns if GPU is available.
        response = requests.post(PREDICT_URL, json={})
        assert response.status_code == 200
        assert response.json() is False


@pytest.mark.integration
@pytest.mark.parametrize("pydantic_major_version", ["1", "2"])
def test_requirements_pydantic(pydantic_major_version):
    with ensure_kill_all():
        truss_root = Path(__file__).parent.parent.parent.resolve() / "truss"
        truss_dir = truss_root / "test_data" / f"test_pyantic_v{pydantic_major_version}"
        tr = TrussHandle(truss_dir)
        _ = tr.docker_run(local_port=8090, detach=True, wait_for_server_ready=True)

        response = requests.post(PREDICT_URL, json={})
        assert response.status_code == 200
        assert response.json() == '{\n    "foo": "bla",\n    "bar": 123\n}'


@pytest.mark.integration
def test_async_truss():
    with ensure_kill_all():
        truss_root = Path(__file__).parent.parent.parent.resolve() / "truss"
        truss_dir = truss_root / "test_data" / "test_async_truss"
        tr = TrussHandle(truss_dir)
        _ = tr.docker_run(local_port=8090, detach=True, wait_for_server_ready=True)

        response = requests.post(PREDICT_URL, json={})
        assert response.json() == {
            "preprocess_value": "value",
            "postprocess_value": "value",
        }


@pytest.mark.integration
def test_async_streaming():
    with ensure_kill_all():
        truss_root = Path(__file__).parent.parent.parent.resolve() / "truss"
        truss_dir = truss_root / "test_data" / "test_streaming_async_generator_truss"
        tr = TrussHandle(truss_dir)
        _ = tr.docker_run(local_port=8090, detach=True, wait_for_server_ready=True)

        response = requests.post(PREDICT_URL, json={}, stream=True)
        assert response.headers.get("transfer-encoding") == "chunked"
        assert [
            byte_string.decode() for byte_string in list(response.iter_content())
        ] == ["0", "1", "2", "3", "4"]

        predict_non_stream_response = requests.post(
            PREDICT_URL,
            json={},
            stream=True,
            headers={"accept": "application/json"},
        )
        assert "transfer-encoding" not in predict_non_stream_response.headers
        assert predict_non_stream_response.json() == "01234"


@pytest.mark.integration
def test_async_streaming_timeout():
    with ensure_kill_all():
        truss_root = Path(__file__).parent.parent.parent.resolve() / "truss"
        truss_dir = truss_root / "test_data" / "test_streaming_read_timeout"
        tr = TrussHandle(truss_dir)
        container = tr.docker_run(
            local_port=8090, detach=True, wait_for_server_ready=True
        )

        # ChunkedEncodingError is raised when the chunk does not get processed due to streaming read timeout
        with pytest.raises(requests.exceptions.ChunkedEncodingError):
            response = requests.post(PREDICT_URL, json={}, stream=True)

            for chunk in response.iter_content():
                pass

        # Check to ensure the Timeout error is in the container logs
        # TODO: maybe intercept this error better?
        assert_logs_contain_error(
            container.logs(),
            error="raise exceptions.TimeoutError()",
            message="Exception in ASGI application\n",
        )


@pytest.mark.integration
def test_streaming_with_error_and_stacktrace():
    with ensure_kill_all():
        truss_root = Path(__file__).parent.parent.parent.resolve() / "truss"
        truss_dir = truss_root / "test_data" / "test_streaming_truss_with_error"
        tr = TrussHandle(truss_dir)
        container = tr.docker_run(
            local_port=8090, detach=True, wait_for_server_ready=True
        )

        predict_error_response = requests.post(
            PREDICT_URL, json={"throw_error": True}, stream=True, timeout=2
        )

        # In error cases, the response will return whatever the stream returned,
        # in this case, the first 3 items. We timeout after 2 seconds to ensure that
        # stream finishes reading and releases the predict semaphore.
        assert [
            byte_string.decode()
            for byte_string in predict_error_response.iter_content()
        ] == ["0", "1", "2"]

        # Test that we are able to continue to make requests successfully
        predict_non_error_response = requests.post(
            PREDICT_URL, json={"throw_error": False}, stream=True, timeout=2
        )

        assert [
            byte_string.decode()
            for byte_string in predict_non_error_response.iter_content()
        ] == [
            "0",
            "1",
            "2",
            "3",
            "4",
        ]
        expected_stack_trace = (
            "Traceback (most recent call last):\n"
            '  File "/app/model/model.py", line 12, in inner\n'
            "    helpers_1.foo(123)\n"
            '  File "/packages/helpers_1.py", line 5, in foo\n'
            "    return helpers_2.bar(x)\n"
            '  File "/packages/helpers_2.py", line 2, in bar\n'
            '    raise Exception("Crashed in `bar`.")\n'
            "Exception: Crashed in `bar`."
        )
        assert_logs_contain_error(
            container.logs(),
            error=expected_stack_trace,
            message="Exception while generating streamed response: Crashed in `bar`.",
        )


@pytest.mark.integration
def test_secrets_truss():
    class Model:
        def __init__(self, **kwargs):
            self._secrets = kwargs["secrets"]

        def predict(self, request):
            return self._secrets["secret"]

    config = """model_name: secrets-truss
secrets:
    secret: null
    """

    config_with_no_secret = "model_name: secrets-truss"
    missing_secret_error_message = """Secret 'secret' not found. Please ensure that:
  * Secret 'secret' is defined in the 'secrets' section of the Truss config file
  * The model was pushed with the --trusted flag"""

    with ensure_kill_all(), temp_truss(inspect.getsource(Model), config) as tr:
        LocalConfigHandler.set_secret("secret", "secret_value")
        _ = tr.docker_run(local_port=8090, detach=True, wait_for_server_ready=True)

        response = requests.post(PREDICT_URL, json={})

        assert response.json() == "secret_value"

    # Case where the secret is not specified in the config
    with ensure_kill_all(), temp_truss(
        inspect.getsource(Model), config_with_no_secret
    ) as tr:
        LocalConfigHandler.set_secret("secret", "secret_value")
        container = tr.docker_run(
            local_port=8090, detach=True, wait_for_server_ready=True
        )

        response = requests.post(PREDICT_URL, json={})
        assert "error" in response.json()
        assert_logs_contain_error(container.logs(), missing_secret_error_message)
        assert "Internal Server Error" in response.json()["error"]
        assert response.headers["x-baseten-error-source"] == "04"
        assert response.headers["x-baseten-error-code"] == "600"

    # Case where the secret is not mounted
    with ensure_kill_all(), temp_truss(inspect.getsource(Model), config) as tr:
        LocalConfigHandler.remove_secret("secret")
        container = tr.docker_run(
            local_port=8090, detach=True, wait_for_server_ready=True
        )

        response = requests.post(PREDICT_URL, json={})
        assert response.status_code == 500
        assert_logs_contain_error(container.logs(), missing_secret_error_message)
        assert "Internal Server Error" in response.json()["error"]
        assert response.headers["x-baseten-error-source"] == "04"
        assert response.headers["x-baseten-error-code"] == "600"


@pytest.mark.integration
def test_postprocess_with_streaming_predict():
    # TODO: revisit the decision to forbid this. If so remove below comment.
    """
    Test a Truss that has streaming response from both predict and postprocess.
    In this case, the postprocess step continues to happen within the predict lock,
    so we don't bother testing the lock scenario, just the behavior that the postprocess
    function is applied.
    """
    model = """
    import time

    class Model:
        def postprocess(self, response):
            for item in response:
                time.sleep(1)
                yield item + " modified"

        def predict(self, request):
            for i in range(2):
                yield str(i)
    """

    config = "model_name: error-truss"
    with ensure_kill_all(), temp_truss(model, config) as tr:
        container = tr.docker_run(
            local_port=8090, detach=True, wait_for_server_ready=True
        )

        response = requests.post(PREDICT_URL, json={}, stream=True)
        logging.info(response.content)
        assert_logs_contain_error(
            container.logs(),
            "ModelDefinitionError: If the predict function returns a generator (streaming), you cannot use postprocessing.",
        )
        assert "Internal Server Error" in response.json()["error"]
        assert response.headers["x-baseten-error-source"] == "04"
        assert response.headers["x-baseten-error-code"] == "600"


@pytest.mark.integration
def test_streaming_postprocess():
    """
    Tests a Truss where predict returns non-streaming, but postprocess is streamed, and
    ensures that the postprocess step does not happen within the predict lock. To do this,
    we sleep for two seconds during the postprocess streaming process, and fire off two
    requests with a total timeout of 3 seconds, ensuring that if they were serialized
    the test would fail.
    """
    model = """
    import time

    class Model:
        def postprocess(self, response):
            for item in response:
                time.sleep(1)
                yield item + " modified"

        def predict(self, request):
            return ["0", "1"]
    """

    config = "model_name: streaming-truss"
    with ensure_kill_all(), temp_truss(model, config) as tr:
        _ = tr.docker_run(local_port=8090, detach=True, wait_for_server_ready=True)

        def make_request(delay: int):
            # For streamed responses, requests does not start receiving content from server until
            # `iter_content` is called, so we must call this in order to get an actual timeout.
            time.sleep(delay)
            response = requests.post(PREDICT_URL, json={}, stream=True)

            assert response.status_code == 200
            assert response.content == b"0 modified1 modified"

        with ThreadPoolExecutor() as e:
            # We use concurrent.futures.wait instead of the timeout property
            # on requests, since requests timeout property has a complex interaction
            # with streaming.
            first_request = e.submit(make_request, 0)
            second_request = e.submit(make_request, 0.2)
            futures = [first_request, second_request]
            done, _ = concurrent.futures.wait(futures, timeout=3)
            # Ensure that both requests complete within the 3 second timeout,
            # as the predict lock is not held through the postprocess step
            assert first_request in done
            assert second_request in done

            for future in done:
                # Ensure that both futures completed without error
                future.result()


@pytest.mark.integration
def test_postprocess():
    """
    Tests a Truss that has a postprocess step defined, and ensures that the
    postprocess does not happen within the predict lock. To do this, we sleep
    for two seconds during the postprocess, and fire off two requests with a total
    timeout of 3 seconds, ensureing that if they were serialized the test would fail.
    """

    model = """
    import time

    class Model:
        def postprocess(self, response):
            updated_items = []
            for item in response:
                time.sleep(1)
                updated_items.append(item + " modified")
            return updated_items

        def predict(self, request):
            return ["0", "1"]

    """

    config = "model_name: postprocess-truss"
    with ensure_kill_all(), temp_truss(model, config) as tr:
        _ = tr.docker_run(local_port=8090, detach=True, wait_for_server_ready=True)

        def make_request(delay: int):
            time.sleep(delay)
            response = requests.post(PREDICT_URL, json={})
            assert response.status_code == 200
            assert response.json() == ["0 modified", "1 modified"]

        with ThreadPoolExecutor() as e:
            # We use concurrent.futures.wait instead of the timeout property
            # on requests, since requests timeout property has a complex interaction
            # with streaming.
            first_request = e.submit(make_request, 0)
            second_request = e.submit(make_request, 0.2)
            futures = [first_request, second_request]
            done, _ = concurrent.futures.wait(futures, timeout=3)
            # Ensure that both requests complete within the 3 second timeout,
            # as the predict lock is not held through the postprocess step
            assert first_request in done
            assert second_request in done

            for future in done:
                # Ensure that both futures completed without error
                future.result()


@pytest.mark.integration
def test_truss_with_errors():
    model = """
    class Model:
        def predict(self, request):
            raise ValueError("error")
    """

    config = "model_name: error-truss"

    with ensure_kill_all(), temp_truss(model, config) as tr:
        container = tr.docker_run(
            local_port=8090, detach=True, wait_for_server_ready=True
        )

        response = requests.post(PREDICT_URL, json={})
        assert response.status_code == 500
        assert "error" in response.json()

        assert_logs_contain_error(container.logs(), "ValueError: error")

        assert "Internal Server Error" in response.json()["error"]
        assert response.headers["x-baseten-error-source"] == "04"
        assert response.headers["x-baseten-error-code"] == "600"

    model_preprocess_error = """
    class Model:
        def preprocess(self, request):
            raise ValueError("error")

        def predict(self, request):
            return {"a": "b"}
    """

    with ensure_kill_all(), temp_truss(model_preprocess_error, config) as tr:
        container = tr.docker_run(
            local_port=8090, detach=True, wait_for_server_ready=True
        )

        response = requests.post(PREDICT_URL, json={})
        assert response.status_code == 500
        assert "error" in response.json()

        assert_logs_contain_error(container.logs(), "ValueError: error")
        assert "Internal Server Error" in response.json()["error"]
        assert response.headers["x-baseten-error-source"] == "04"
        assert response.headers["x-baseten-error-code"] == "600"

    model_postprocess_error = """
    class Model:
        def predict(self, request):
            return {"a": "b"}

        def postprocess(self, response):
            raise ValueError("error")
    """

    with ensure_kill_all(), temp_truss(model_postprocess_error, config) as tr:
        container = tr.docker_run(
            local_port=8090, detach=True, wait_for_server_ready=True
        )

        response = requests.post(PREDICT_URL, json={})
        assert response.status_code == 500
        assert "error" in response.json()
        assert_logs_contain_error(container.logs(), "ValueError: error")
        assert "Internal Server Error" in response.json()["error"]
        assert response.headers["x-baseten-error-source"] == "04"
        assert response.headers["x-baseten-error-code"] == "600"

    model_async = """
    class Model:
        async def predict(self, request):
            raise ValueError("error")
    """

    with ensure_kill_all(), temp_truss(model_async, config) as tr:
        container = tr.docker_run(
            local_port=8090, detach=True, wait_for_server_ready=True
        )

        response = requests.post(PREDICT_URL, json={})
        assert response.status_code == 500
        assert "error" in response.json()

        assert_logs_contain_error(container.logs(), "ValueError: error")

        assert "Internal Server Error" in response.json()["error"]
        assert response.headers["x-baseten-error-source"] == "04"
        assert response.headers["x-baseten-error-code"] == "600"


@pytest.mark.integration
def test_truss_with_user_errors():
    """Test that user-code raised `fastapi.HTTPExceptions` are passed through as is."""
    model = """
    import fastapi

    class Model:
        def predict(self, request):
            raise fastapi.HTTPException(status_code=500, detail="My custom message.")
    """

    config = "model_name: error-truss"

    with ensure_kill_all(), temp_truss(model, config) as tr:
        container = tr.docker_run(
            local_port=8090, detach=True, wait_for_server_ready=True
        )

        response = requests.post(PREDICT_URL, json={})
        assert response.status_code == 500
        assert "error" in response.json()
        assert response.headers["x-baseten-error-source"] == "04"
        assert response.headers["x-baseten-error-code"] == "600"

        assert_logs_contain_error(
            container.logs(),
            "HTTPException: 500: My custom message.",
            "Model raised HTTPException",
        )

        assert "My custom message." in response.json()["error"]
        assert response.headers["x-baseten-error-source"] == "04"
        assert response.headers["x-baseten-error-code"] == "600"


@pytest.mark.integration
def test_truss_with_error_stacktrace():
    with ensure_kill_all():
        truss_root = Path(__file__).parent.parent.parent.resolve() / "truss"
        truss_dir = truss_root / "test_data" / "test_truss_with_error"
        tr = TrussHandle(truss_dir)
        container = tr.docker_run(
            local_port=8090, detach=True, wait_for_server_ready=True
        )

        response = requests.post(PREDICT_URL, json={})
        assert response.status_code == 500
        assert "error" in response.json()

        assert "Internal Server Error" in response.json()["error"]
        assert response.headers["x-baseten-error-source"] == "04"
        assert response.headers["x-baseten-error-code"] == "600"

        expected_stack_trace = (
            "Traceback (most recent call last):\n"
            '  File "/app/model/model.py", line 8, in predict\n'
            "    return helpers_1.foo(123)\n"
            '  File "/packages/helpers_1.py", line 5, in foo\n'
            "    return helpers_2.bar(x)\n"
            '  File "/packages/helpers_2.py", line 2, in bar\n'
            '    raise Exception("Crashed in `bar`.")\n'
            "Exception: Crashed in `bar`."
        )
        assert_logs_contain_error(
            container.logs(),
            error=expected_stack_trace,
            message="Internal Server Error",
        )


@pytest.mark.integration
def test_slow_truss():
    with ensure_kill_all():
        truss_root = Path(__file__).parent.parent.parent.resolve() / "truss"
        truss_dir = truss_root / "test_data" / "server_conformance_test_truss"
        tr = TrussHandle(truss_dir)

        _ = tr.docker_run(local_port=8090, detach=True, wait_for_server_ready=False)

        truss_server_addr = "http://localhost:8090"

        def _test_liveness_probe(expected_code):
            live = requests.get(f"{truss_server_addr}/")
            assert live.status_code == expected_code

        def _test_readiness_probe(expected_code):
            ready = requests.get(f"{truss_server_addr}/v1/models/model")
            assert ready.status_code == expected_code

        def _test_ping(expected_code):
            ping = requests.get(f"{truss_server_addr}/ping")
            assert ping.status_code == expected_code

        def _test_invocations(expected_code):
            invocations = requests.post(f"{truss_server_addr}/invocations", json={})
            assert invocations.status_code == expected_code

        SERVER_WARMUP_TIME = 3
        LOAD_TEST_TIME = 12
        LOAD_BUFFER_TIME = 7
        PREDICT_TEST_TIME = 15

        # Sleep a few seconds to get the server some time to wake up
        time.sleep(SERVER_WARMUP_TIME)

        # The truss takes about 30 seconds to load.
        # We want to make sure that it's not ready for that time.
        for _ in range(LOAD_TEST_TIME):
            _test_liveness_probe(200)
            _test_readiness_probe(503)
            _test_ping(503)
            _test_invocations(503)
            time.sleep(1)

        time.sleep(LOAD_BUFFER_TIME)
        _test_liveness_probe(200)
        _test_readiness_probe(200)
        _test_ping(200)

        predict_call = Thread(
            target=lambda: requests.post(
                f"{truss_server_addr}/v1/models/model:predict", json={}
            )
        )
        predict_call.start()

        for _ in range(PREDICT_TEST_TIME):
            _test_liveness_probe(200)
            _test_readiness_probe(200)
            _test_ping(200)
            time.sleep(1)

        predict_call.join()

        _test_invocations(200)


# Tracing ##############################################################################


def _make_otel_headers() -> Mapping[str, str]:
    """
    Create and return a mapping with OpenTelemetry trace context headers.

    This function starts a new span and injects the trace context into the headers,
    which can be used to propagate tracing information in outgoing HTTP requests.

    Returns:
        Mapping[str, str]: A mapping containing the trace context headers.
    """
    # Initialize a tracer
    tracer = trace.get_tracer(__name__)

    # Create a dictionary to hold the headers
    headers: dict[str, str] = {}

    # Start a new span
    with tracer.start_as_current_span("outgoing-request-span"):
        # Use the TraceContextTextMapPropagator to inject the trace context into the headers
        propagator = tracecontext.TraceContextTextMapPropagator()
        propagator.inject(headers, context=context.get_current())

    return headers


@pytest.mark.integration
@pytest.mark.parametrize("enable_tracing_data", [True, False])
def test_streaming_truss_with_user_tracing(enable_tracing_data):
    with ensure_kill_all():
        truss_root = Path(__file__).parent.parent.parent.resolve() / "truss"
        truss_dir = truss_root / "test_data" / "test_streaming_truss_with_tracing"
        tr = TrussHandle(truss_dir)

        def enable_gpu_fn(conf):
            new_runtime = dataclasses.replace(
                conf.runtime, enable_tracing_data=enable_tracing_data
            )
            return dataclasses.replace(conf, runtime=new_runtime)

        tr._update_config(enable_gpu_fn)

        container = tr.docker_run(
            local_port=8090, detach=True, wait_for_server_ready=True
        )

        # A request for which response is not completely read
        headers_0 = _make_otel_headers()
        predict_response = requests.post(
            PREDICT_URL, json={}, stream=True, headers=headers_0
        )
        # We just read the first part and leave it hanging here
        next(predict_response.iter_content())

        headers_1 = _make_otel_headers()
        predict_response = requests.post(
            PREDICT_URL, json={}, stream=True, headers=headers_1
        )
        assert predict_response.headers.get("transfer-encoding") == "chunked"

        # When accept is set to application/json, the response is not streamed.
        headers_2 = _make_otel_headers()
        predict_non_stream_response = requests.post(
            PREDICT_URL,
            json={},
            stream=True,
            headers={**headers_2, "accept": "application/json"},
        )
        assert "transfer-encoding" not in predict_non_stream_response.headers
        assert predict_non_stream_response.json() == "01234"

        with tempfile.TemporaryDirectory() as tmp_dir:
            truss_traces_file = pathlib.Path(tmp_dir) / "otel_traces.ndjson"
            container.copy_from("/tmp/otel_traces.ndjson", truss_traces_file)
            truss_traces = [
                json.loads(s) for s in truss_traces_file.read_text().splitlines()
            ]

            user_traces_file = pathlib.Path(tmp_dir) / "otel_user_traces.ndjson"
            container.copy_from("/tmp/otel_user_traces.ndjson", user_traces_file)
            user_traces = [
                json.loads(s) for s in user_traces_file.read_text().splitlines()
            ]

        if not enable_tracing_data:
            assert len(truss_traces) == 0
            assert len(user_traces) > 0
            return

        assert sum(1 for x in truss_traces if x["name"] == "predict-endpoint") == 3
        assert sum(1 for x in user_traces if x["name"] == "load_model") == 1
        assert sum(1 for x in user_traces if x["name"] == "predict") == 3

        user_parents = set(x["parent_id"] for x in user_traces)
        truss_spans = set(x["context"]["span_id"] for x in truss_traces)
        truss_parents = set(x["parent_id"] for x in truss_traces)
        # Make sure there is no context creep into user traces. No user trace should
        # have a truss trace as parent.
        assert user_parents & truss_spans == set()
        # But make sure traces have parents at all.
        assert len(user_parents) > 3
        assert len(truss_parents) > 3


# Returning Response Objects ###########################################################


@pytest.mark.integration
def test_truss_with_response():
    """Test that user-code can set a custom status code."""
    model = """
    from fastapi.responses import Response

    class Model:
        def predict(self, inputs):
            return Response(status_code=inputs["code"])
    """
    from fastapi import status

    config = "model_name: custom-status-code-truss"

    with ensure_kill_all(), temp_truss(model, config) as tr:
        _ = tr.docker_run(local_port=8090, detach=True, wait_for_server_ready=True)

        response = requests.post(PREDICT_URL, json={"code": status.HTTP_204_NO_CONTENT})
        assert response.status_code == 204
        assert "x-baseten-error-source" not in response.headers
        assert "x-baseten-error-code" not in response.headers

        response = requests.post(
            PREDICT_URL, json={"code": status.HTTP_500_INTERNAL_SERVER_ERROR}
        )
        assert response.status_code == 500
        assert response.headers["x-baseten-error-source"] == "04"
        assert response.headers["x-baseten-error-code"] == "700"


@pytest.mark.integration
def test_truss_with_streaming_response():
    # TODO: one issue with this is that (unlike our "builtin" streaming), this keeps
    #  the semaphore claimed potentially longer if the client drops.

    model = """from starlette.responses import StreamingResponse
class Model:
    def predict(self, model_input):
        def text_generator():
            for i in range(3):
                yield f"data: {i}\\n\\n"
        return StreamingResponse(text_generator(), media_type="text/event-stream")
    """

    config = "model_name: sse-truss"

    with ensure_kill_all(), temp_truss(model, config) as tr:
        _ = tr.docker_run(local_port=8090, detach=True, wait_for_server_ready=True)

        # A request for which response is not completely read.
        predict_response = requests.post(PREDICT_URL, json={}, stream=True)
        assert (
            predict_response.headers["Content-Type"]
            == "text/event-stream; charset=utf-8"
        )

        lines = predict_response.text.strip().split("\n")
        assert lines == [
            "data: 0",
            "",
            "data: 1",
            "",
            "data: 2",
        ]


# Using Request in Model ###############################################################


@pytest.mark.integration
def test_truss_with_request():
    model = """
    import fastapi
    class Model:
        async def preprocess(self, request: fastapi.Request):
            return await request.json()

        async def predict(self, inputs, request: fastapi.Request):
            inputs["request_size"] = len(await request.body())
            return inputs

        def postprocess(self, inputs):
             return {**inputs, "postprocess": "was here"}
    """
    with ensure_kill_all(), temp_truss(model, "") as tr:
        _ = tr.docker_run(local_port=8090, detach=True, wait_for_server_ready=True)

        response = requests.post(PREDICT_URL, json={"test": 123})
        assert response.status_code == 200
        assert response.json() == {
            "test": 123,
            "request_size": 13,
            "postprocess": "was here",
        }


@pytest.mark.integration
def test_truss_with_requests_and_invalid_signatures():
    model = """
    class Model:
        def predict(self, inputs, invalid_arg): ...
    """
    with ensure_kill_all(), temp_truss(model, "") as tr:
        container = tr.docker_run(
            local_port=8090, detach=True, wait_for_server_ready=False
        )
        time.sleep(1.0)  # Wait for logs.
        assert_logs_contain_error(
            container.logs(),
            "`predict` method with two arguments must have request as second argument",
            "Exception while loading model",
        )

    model = """
    import fastapi

    class Model:
        def predict(self, request: fastapi.Request, invalid_arg): ...
    """
    with ensure_kill_all(), temp_truss(model, "") as tr:
        container = tr.docker_run(
            local_port=8090, detach=True, wait_for_server_ready=False
        )
        time.sleep(1.0)  # Wait for logs.
        assert_logs_contain_error(
            container.logs(),
            "`predict` method with two arguments is not allowed to have request as "
            "first argument",
            "Exception while loading model",
        )

    model = """
    import fastapi

    class Model:
        def predict(self, inputs, request: fastapi.Request, something): ...
    """
    with ensure_kill_all(), temp_truss(model, "") as tr:
        container = tr.docker_run(
            local_port=8090, detach=True, wait_for_server_ready=False
        )
        time.sleep(1.0)  # Wait for logs.
        assert_logs_contain_error(
            container.logs(),
            "`predict` method cannot have more than two arguments",
            "Exception while loading model",
        )


@pytest.mark.integration
def test_truss_with_requests_and_invalid_argument_combinations():
    model = """
    import fastapi
    class Model:
        async def preprocess(self, inputs): ...

        def predict(self, request: fastapi.Request): ...
    """
    with ensure_kill_all(), temp_truss(model, "") as tr:
        container = tr.docker_run(
            local_port=8090, detach=True, wait_for_server_ready=False
        )
        time.sleep(1.0)  # Wait for logs.
        assert_logs_contain_error(
            container.logs(),
            "When using preprocessing, the predict method cannot only have the request argument",
            "Exception while loading model",
        )

    model = """
    import fastapi
    class Model:
        def preprocess(self, inputs): ...

        async def predict(self, inputs, request: fastapi.Request): ...

        def postprocess(self, request: fastapi.Request): ...
    """
    with ensure_kill_all(), temp_truss(model, "") as tr:
        container = tr.docker_run(
            local_port=8090, detach=True, wait_for_server_ready=False
        )
        time.sleep(1.0)  # Wait for logs.
        assert_logs_contain_error(
            container.logs(),
            "The postprocessing method cannot only have the request argument",
            "Exception while loading model",
        )

    model = """
    import fastapi
    class Model:
        def preprocess(self, inputs): ...
    """
    with ensure_kill_all(), temp_truss(model, "") as tr:
        container = tr.docker_run(
            local_port=8090, detach=True, wait_for_server_ready=False
        )
        time.sleep(1.0)  # Wait for logs.
        assert_logs_contain_error(
            container.logs(),
            "Truss model must have a `predict` method.",
            "Exception while loading model",
        )


@pytest.mark.integration
def test_truss_forbid_postprocessing_with_response():
    model = """
    import fastapi, json
    class Model:
        def predict(self, inputs):
            return fastapi.Response(content=json.dumps(inputs), status_code=200)

        def postprocess(self, inputs):
             return inputs
    """
    with ensure_kill_all(), temp_truss(model, "") as tr:
        container = tr.docker_run(
            local_port=8090, detach=True, wait_for_server_ready=True
        )

        response = requests.post(PREDICT_URL, json={})
        assert response.status_code == 500
        assert response.headers["x-baseten-error-source"] == "04"
        assert response.headers["x-baseten-error-code"] == "600"
        assert_logs_contain_error(
            container.logs(),
            "If the predict function returns a response object, you cannot "
            "use postprocessing.",
        )


@pytest.mark.integration
def test_async_streaming_with_cancellation():
    model = """
    import fastapi, asyncio, logging

    class Model:
        async def predict(self, inputs, request: fastapi.Request):
            await asyncio.sleep(1)
            if await request.is_disconnected():
                logging.warning("Cancelled (before gen).")
                return

            for i in range(5):
                await asyncio.sleep(1.0)
                logging.warning(i)
                yield str(i)
                if await request.is_disconnected():
                    logging.warning("Cancelled (during gen).")
                    return
    """
    with ensure_kill_all(), temp_truss(model, "") as tr:
        container = tr.docker_run(
            local_port=8090, detach=True, wait_for_server_ready=True
        )
        # For hard cancellation we need to use httpx, requests' timeouts don't work.
        with pytest.raises(httpx.ReadTimeout):
            with httpx.Client(
                timeout=httpx.Timeout(1.0, connect=1.0, read=1.0)
            ) as client:
                response = client.post(PREDICT_URL, json={}, timeout=1.0)
                response.raise_for_status()

        time.sleep(2)  # Wait a bit to get all logs.
        assert "Cancelled (during gen)." in container.logs()


@pytest.mark.integration
def test_limit_concurrency_with_sse():
    # It seems that the "builtin" functionality of the FastAPI server already buffers
    # the generator, so that it doesn't keep hanging around if the client doesn't
    # consume data. `_buffered_response_generator` might be redundant.
    # This can be observed by waiting for a long time in `make_request`: the server will
    # print `Done` for the tasks, while we still wait and hold the unconsumed response.
    # For testing we need to have actually slow generation to keep the server busy.
    model = """
    import asyncio

    class Model:
        async def predict(self, request):
            print(f"Starting {request}")
            for i in range(5):
                await asyncio.sleep(0.1)
                yield str(i)
            print(f"Done {request}")

    """

    config = """runtime:
  predict_concurrency: 2"""

    def make_request(consume_chunks, timeout, task_id):
        t0 = time.time()
        with httpx.Client() as client:
            with client.stream(
                "POST", PREDICT_URL, json={"task_id": task_id}
            ) as response:
                assert response.status_code == 200
                if consume_chunks:
                    chunks = [chunk for chunk in response.iter_text()]
                    print(f"consumed chunks ({task_id}): {chunks}")
                    assert len(chunks) > 0
                    t1 = time.time()
                    if t1 - t0 > timeout:
                        raise httpx.ReadTimeout("Timeout")
                    return chunks
                else:
                    print(f"waiting ({task_id})")
                    time.sleep(0.5)  # Hold the connection.
                    print(f"waiting done ({task_id})")

    with ensure_kill_all(), temp_truss(model, config) as tr:
        _ = tr.docker_run(local_port=8090, detach=True, wait_for_server_ready=True)
        # Processing full request takes 0.5s.
        print("Make warmup request")
        make_request(consume_chunks=True, timeout=0.55, task_id=0)

        with ThreadPoolExecutor() as executor:
            # Start two requests and hold them without consuming all chunks
            # Each takes for 0.5 s. Semaphore should be claimed, with 0 remaining.
            print("Start two tasks.")
            task1 = executor.submit(make_request, False, 0.55, 1)
            task2 = executor.submit(make_request, False, 0.55, 2)
            print("Wait for tasks to start.")
            time.sleep(0.05)
            print("Make a request while server is busy.")
            with pytest.raises(httpx.ReadTimeout):
                make_request(True, timeout=0.55, task_id=3)

            task1.result()
            task2.result()
            print("Task 1 and 2 completed. Server should be free again.")

        result = make_request(True, timeout=0.55, task_id=4)
        print(f"Final chunks: {result}")



import shutil
from pathlib import Path

import requests
from truss.blob.blob_backend import BlobBackend

BLOB_DOWNLOAD_TIMEOUT_SECS = 600  # 10 minutes


class HttpPublic(BlobBackend):
    """Downloads without auth, files must be publicly available."""

    def download(self, URL: str, download_to: Path):
        # Streaming download to keep memory usage low
        resp = requests.get(
            URL,
            allow_redirects=True,
            stream=True,
            timeout=BLOB_DOWNLOAD_TIMEOUT_SECS,
        )
        resp.raise_for_status()
        with download_to.open("wb") as file:
            shutil.copyfileobj(resp.raw, file)



import os
import random
import shutil
import string
import tempfile
from contextlib import contextmanager
from pathlib import Path
from typing import Iterable, Iterator, List, Optional, Set, Tuple, Union

import pathspec

# .truss_ignore is a fixed file in the Truss library that is used to specify files
# that should be ignored when copying a directory tree such as .git directory.
FIXED_TRUSS_IGNORE_PATH = Path(__file__).parent / ".truss_ignore"


def copy_tree_path(src: Path, dest: Path, ignore_patterns: List[str] = []) -> None:
    """Copy a directory tree, ignoring files specified in .truss_ignore."""
    patterns = load_trussignore_patterns()
    patterns.extend(ignore_patterns)

    if not dest.exists():
        dest.mkdir(parents=True)

    for sub_path in src.rglob("*"):
        if is_ignored(sub_path, patterns, base_dir=src):
            continue

        dest_fp = dest / sub_path.relative_to(src)

        if sub_path.is_dir():
            dest_fp.mkdir(exist_ok=True)
        else:
            dest_fp.parent.mkdir(exist_ok=True, parents=True)
            shutil.copy2(str(sub_path), str(dest_fp))


def copy_file_path(src: Path, dest: Path) -> Tuple[str, str]:
    return shutil.copy2(str(src), str(dest))


def copy_tree_or_file(src: Path, dest: Path) -> Union[List[str], Tuple[str, str]]:
    if src.is_file():
        return copy_file_path(src, dest)

    return copy_tree_path(src, dest)  # type: ignore


def remove_tree_path(target: Path) -> None:
    return shutil.rmtree(str(target))


def get_max_modified_time_of_dir(path: Path) -> float:
    max_modified_time = os.path.getmtime(path)
    for root, dirs, files in os.walk(path):
        if os.path.islink(root):
            raise ValueError(f"Symlinks not allowed in Truss: {root}")
        files = [f for f in files if not f.startswith(".")]
        dirs[:] = [d for d in dirs if not d.startswith(".")]
        max_modified_time = max(max_modified_time, os.path.getmtime(root))
        for file in files:
            max_modified_time = max(
                max_modified_time, os.path.getmtime(os.path.join(root, file))
            )
    return max_modified_time


@contextmanager
def given_or_temporary_dir(given_dir: Optional[Path] = None):
    if given_dir is not None:
        yield given_dir
    else:
        with tempfile.TemporaryDirectory() as temp_dir:
            yield Path(temp_dir)


def build_truss_target_directory(stub: str) -> Path:
    """Builds a directory under ~/.truss/models for the purpose of creating a Truss at."""
    rand_suffix = "".join(random.choices(string.ascii_uppercase + string.digits, k=6))
    target_directory_path = Path(
        Path.home(), ".truss", "models", f"{stub}-{rand_suffix}"
    )
    target_directory_path.mkdir(parents=True)
    return target_directory_path


def load_trussignore_patterns(
    truss_ignore_file: Path = FIXED_TRUSS_IGNORE_PATH,
) -> List[str]:
    """Load patterns from a .truss_ignore file"""
    patterns = []

    with truss_ignore_file.open() as f:
        lines = f.readlines()

    for line in lines:
        line = line.strip()
        if line and not line.startswith("#"):
            patterns.append(line)

    return patterns


def is_ignored(
    path: Path, patterns: List[str], base_dir: Optional[Path] = None
) -> bool:
    """
    Determines if a specified path (or any of its parent directories) should be ignored
    based on a list of ignore patterns, analogous to how .gitignore works in Git. The
    function evaluates not only the path itself but all its parent directories up to
    the base directory. Hence, if a file resides within a directory to be ignored, it
    will be ignored as well. We provide the 'base_dir' argument to allow the caller to
    specify the base directory against which the relative 'path' is resolved. This
    prevents false positives where the absolute path before the base directory matches
    one of the ignore patterns.

    Args:
        path (Path): The path to the file or directory being checked. This can be absolute
            or relative. If 'base_dir' is provided, 'path' is considered relative to 'base_dir'.
        patterns (List[str]): A list of ignore patterns. Each pattern is a string that can contain
            wildcard characters. For instance, '*.py' would match all Python files, while 'test/'
            would match a directory named 'test'. A pattern ending with a slash (/) will only match directories.
        base_dir (Optional[Path]): If specified, it's used as the base directory against which
            the relative 'path' is resolved. If None, the 'path' is treated as either an absolute path
            or relative to the current working directory.

    Returns:
        bool: True if the path matches any of the ignore patterns (i.e., should be ignored),
            and False otherwise.
    """
    ignore_spec = pathspec.PathSpec.from_lines(
        pathspec.patterns.GitWildMatchPattern, patterns
    )

    if base_dir:
        path = path.relative_to(base_dir)

    return ignore_spec.match_file(path)


def get_ignored_relative_paths(
    root_relative_paths: Iterable[Union[str, os.PathLike]],
    ignore_patterns: Optional[List[str]] = None,
) -> Iterator[Union[str, os.PathLike]]:
    """Given an iterable of relative paths, returns an iterator of the relative paths that match ignore_patterns."""
    if ignore_patterns is None:
        return iter([])

    ignore_spec = pathspec.PathSpec.from_lines(
        pathspec.patterns.GitWildMatchPattern, ignore_patterns
    )
    return ignore_spec.match_files(root_relative_paths)


def get_unignored_relative_paths_from_root(
    root: Path,
    ignore_patterns: Optional[List[str]] = None,
) -> Set[Path]:
    """Given a root directory, returns an iterator of the relative paths that do not match ignore_patterns."""
    root_relative_paths = set(path.relative_to(root) for path in root.glob("**/*"))

    ignored_paths = set(
        get_ignored_relative_paths(root_relative_paths, ignore_patterns)
    )
    return root_relative_paths - ignored_paths  # type: ignore



import json
import uuid
from datetime import date, datetime, time, timedelta
from decimal import Decimal
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    Callable,
    Dict,
    Generator,
    List,
    Optional,
    Union,
)

import pydantic
import starlette.responses

if TYPE_CHECKING:
    from numpy.typing import NDArray


JSONType = Union[str, int, float, bool, None, List["JSONType"], Dict[str, "JSONType"]]
MsgPackType = Union[
    str,
    int,
    float,
    bool,
    None,
    date,
    Decimal,
    datetime,
    time,
    timedelta,
    uuid.UUID,
    "NDArray",
    List["MsgPackType"],
    Dict[str, "MsgPackType"],
]
InputType = Union[JSONType, MsgPackType, pydantic.BaseModel]
OutputType = Union[
    JSONType,
    MsgPackType,
    Generator[bytes, None, None],
    AsyncGenerator[bytes, None],
    starlette.responses.Response,
]


# mostly cribbed from django.core.serializer.DjangoJSONEncoder
def _truss_msgpack_encoder(
    obj: Union[Decimal, date, time, timedelta, uuid.UUID, Dict],
    chain: Optional[Callable] = None,
) -> Dict:
    if isinstance(obj, datetime):
        r = obj.isoformat()
        if r.endswith("+00:00"):
            r = r[:-6] + "Z"
        return {b"__dt_datetime_iso__": True, b"data": r}
    elif isinstance(obj, date):
        r = obj.isoformat()
        return {b"__dt_date_iso__": True, b"data": r}
    elif isinstance(obj, time):
        if obj.utcoffset() is not None:
            raise ValueError("Cannot represent timezone-aware times.")
        r = obj.isoformat()
        return {b"__dt_time_iso__": True, b"data": r}
    elif isinstance(obj, timedelta):
        return {
            b"__dt_timedelta__": True,
            b"data": (obj.days, obj.seconds, obj.microseconds),
        }
    elif isinstance(obj, Decimal):
        return {b"__decimal__": True, b"data": str(obj)}
    elif isinstance(obj, uuid.UUID):
        return {b"__uuid__": True, b"data": str(obj)}
    else:
        return obj if chain is None else chain(obj)


def _truss_msgpack_decoder(obj: Any, chain=None):
    try:
        if b"__dt_datetime_iso__" in obj:
            return datetime.fromisoformat(obj[b"data"])
        elif b"__dt_date_iso__" in obj:
            return date.fromisoformat(obj[b"data"])
        elif b"__dt_time_iso__" in obj:
            return time.fromisoformat(obj[b"data"])
        elif b"__dt_timedelta__" in obj:
            days, seconds, microseconds = obj[b"data"]
            return timedelta(days=days, seconds=seconds, microseconds=microseconds)
        elif b"__decimal__" in obj:
            return Decimal(obj[b"data"])
        elif b"__uuid__" in obj:
            return uuid.UUID(obj[b"data"])
        else:
            return obj if chain is None else chain(obj)
    except KeyError:
        return obj if chain is None else chain(obj)


# this json object is JSONType + np.array + datetime
def is_truss_serializable(obj: Any) -> bool:
    import numpy as np

    # basic JSON types
    if isinstance(obj, (str, int, float, bool, type(None), dict, list)):
        return True
    elif isinstance(obj, (datetime, date, time, timedelta)):
        return True
    elif isinstance(obj, np.ndarray):
        return True
    else:
        return False


def truss_msgpack_serialize(obj: MsgPackType) -> bytes:
    import msgpack
    import msgpack_numpy as mp_np

    return msgpack.packb(
        obj, default=lambda x: _truss_msgpack_encoder(x, chain=mp_np.encode)
    )


def truss_msgpack_deserialize(data: bytes) -> MsgPackType:
    import msgpack
    import msgpack_numpy as mp_np

    return msgpack.unpackb(
        data, object_hook=lambda x: _truss_msgpack_decoder(x, chain=mp_np.decode)
    )


class DeepNumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        import numpy as np

        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super().default(obj)



import logging
import subprocess
from pathlib import Path
from typing import List

from truss.templates.control.control.helpers.custom_types import (
    Action,
    ModelCodePatch,
    Patch,
    PythonRequirementPatch,
    SystemPackagePatch,
)
from truss.templates.control.control.helpers.errors import UnsupportedPatch
from truss.templates.control.control.helpers.truss_patch.model_code_patch_applier import (
    apply_code_patch,
)
from truss.truss_config import TrussConfig


class LocalTrussPatchApplier:
    """Applies patches to a truss server running locally.
    This should be compatible with ModelContainerPatchApplier.
    """

    def __init__(self, truss_dir: Path, env_exe, logger: logging.Logger) -> None:
        self._truss_dir = truss_dir
        self._env_exe = env_exe
        self._logger = logger

    def __call__(self, patches: List[Patch]):
        for patch in patches:
            self._logger.debug(f"Applying patch {patch.to_dict()}")
            if isinstance(patch.body, ModelCodePatch):
                model_code_patch: ModelCodePatch = patch.body
                model_module_dir = self._truss_dir / self._truss_config.model_module_dir
                apply_code_patch(model_module_dir, model_code_patch, self._logger)
            elif isinstance(patch.body, PythonRequirementPatch):
                py_req_patch: PythonRequirementPatch = patch.body
                self._apply_python_requirement_patch(py_req_patch)
            elif isinstance(patch.body, SystemPackagePatch):
                self._logger.info(
                    "System package patches are not supported for local server"
                )
            else:
                raise UnsupportedPatch(f"Unknown patch type {patch.type}")

    @property
    def _truss_config(self) -> TrussConfig:
        return TrussConfig.from_yaml(self._truss_dir / "config.yaml")

    def _apply_python_requirement_patch(
        self, python_requirement_patch: PythonRequirementPatch
    ):
        self._logger.debug(
            f"Applying python requirement patch {python_requirement_patch.to_dict()}"
        )
        action = python_requirement_patch.action

        if action == Action.REMOVE:
            subprocess.run(
                [
                    self._env_exe,
                    "-m",
                    "pip",
                    "uninstall",
                    "-y",
                    python_requirement_patch.requirement,
                ],
                check=True,
            )
        elif action in [Action.ADD, Action.UPDATE]:
            subprocess.run(
                [
                    self._env_exe,
                    "-m",
                    "pip",
                    "install",
                    python_requirement_patch.requirement,
                ],
                check=True,
            )
        else:
            raise ValueError(f"Unknown python requirement patch action {action}")



import contextlib
import json
import logging
import os
import pathlib
import time
from typing import Iterator, List, Optional, Sequence

import opentelemetry.exporter.otlp.proto.http.trace_exporter as oltp_exporter
import opentelemetry.sdk.resources as resources
import opentelemetry.sdk.trace as sdk_trace
import opentelemetry.sdk.trace.export as trace_export
from opentelemetry import context, trace
from shared import secrets_resolver

logger = logging.getLogger(__name__)

ATTR_NAME_DURATION = "duration_sec"
# "New" Jaeger exporter.
OTEL_EXPORTER_OTLP_ENDPOINT = "OTEL_EXPORTER_OTLP_ENDPOINT"
# "Old" Honeycomb exporter. We might want to temporarily export to both, but
# eventually remove this one.
OTEL_EXPORTER_OTLP_ENDPOINT_LEGACY = "OTEL_EXPORTER_OTLP_ENDPOINT_LEGACY"
# Writing trace data to a file is only intended for testing / debugging.
OTEL_TRACING_NDJSON_FILE = "OTEL_TRACING_NDJSON_FILE"
# Exporting trace data to a public honeycomb instance (not our cluster collector)
# is intended only for testing / debugging.
HONEYCOMB_DATASET = "HONEYCOMB_DATASET"
HONEYCOMB_API_KEY = "HONEYCOMB_API_KEY"

DEFAULT_ENABLE_TRACING_DATA = False  # This should be in sync with truss_config.py.


class JSONFileExporter(trace_export.SpanExporter):
    """Writes spans to newline-delimited JSON file for debugging / testing."""

    def __init__(self, file_path: pathlib.Path):
        self._file = file_path.open("a")

    def export(
        self, spans: Sequence[sdk_trace.ReadableSpan]
    ) -> trace_export.SpanExportResult:
        for span in spans:
            # Get rid of newlines and whitespace.
            self._file.write(json.dumps(json.loads(span.to_json())))
            self._file.write("\n")
        self._file.flush()
        return trace_export.SpanExportResult.SUCCESS

    def shutdown(self) -> None:
        self._file.close()


_truss_tracer: Optional[trace.Tracer] = None


def get_truss_tracer(secrets: secrets_resolver.Secrets, config) -> trace.Tracer:
    """Creates a cached tracer (i.e. runtime-singleton) to be used for truss
    internal tracing.

    The goal is to separate truss-internal tracing instrumentation
    completely from potential user-defined tracing - see also `detach_context` below.

    """
    enable_tracing_data = config.get("runtime", {}).get(
        "enable_tracing_data", DEFAULT_ENABLE_TRACING_DATA
    )

    global _truss_tracer
    if _truss_tracer:
        return _truss_tracer

    span_processors: List[sdk_trace.SpanProcessor] = []
    if otlp_endpoint := os.getenv(OTEL_EXPORTER_OTLP_ENDPOINT):
        logger.info(f"Exporting trace data to {OTEL_EXPORTER_OTLP_ENDPOINT}.")
        otlp_exporter = oltp_exporter.OTLPSpanExporter(endpoint=otlp_endpoint)
        otlp_processor = sdk_trace.export.BatchSpanProcessor(otlp_exporter)
        span_processors.append(otlp_processor)

    if otlp_endpoint := os.getenv(OTEL_EXPORTER_OTLP_ENDPOINT_LEGACY):
        logger.info(f"Exporting trace data to {OTEL_EXPORTER_OTLP_ENDPOINT_LEGACY}.")
        otlp_exporter = oltp_exporter.OTLPSpanExporter(endpoint=otlp_endpoint)
        otlp_processor = sdk_trace.export.BatchSpanProcessor(otlp_exporter)
        span_processors.append(otlp_processor)

    if tracing_log_file := os.getenv(OTEL_TRACING_NDJSON_FILE):
        logger.info(f"Exporting trace data to file `{tracing_log_file}`.")
        json_file_exporter = JSONFileExporter(pathlib.Path(tracing_log_file))
        file_processor = sdk_trace.export.SimpleSpanProcessor(json_file_exporter)
        span_processors.append(file_processor)

    if (
        honeycomb_dataset := os.getenv(HONEYCOMB_DATASET)
    ) and HONEYCOMB_API_KEY in secrets:
        honeycomb_api_key = secrets[HONEYCOMB_API_KEY]
        logger.info("Exporting trace data to honeycomb.")
        honeycomb_exporter = oltp_exporter.OTLPSpanExporter(
            endpoint="https://api.honeycomb.io/v1/traces",
            headers={
                "x-honeycomb-team": honeycomb_api_key,
                "x-honeycomb-dataset": honeycomb_dataset,
            },
        )
        honeycomb_processor = sdk_trace.export.BatchSpanProcessor(honeycomb_exporter)
        span_processors.append(honeycomb_processor)

    if span_processors and enable_tracing_data:
        logger.info("Instantiating truss tracer.")
        resource = resources.Resource.create({resources.SERVICE_NAME: "TrussServer"})
        trace_provider = sdk_trace.TracerProvider(resource=resource)
        for sp in span_processors:
            trace_provider.add_span_processor(sp)
        tracer = trace_provider.get_tracer("truss_server")
    else:
        if enable_tracing_data:
            logger.info(
                "Using no-op tracing (tracing is enabled, but no exporters configured)."
            )
        else:
            logger.info("Using no-op tracing (tracing was disabled).")

        tracer = sdk_trace.NoOpTracer()

    _truss_tracer = tracer
    return _truss_tracer


@contextlib.contextmanager
def detach_context() -> Iterator[trace.Context]:
    """Breaks opentelemetry's context propagation.

    The goal is to separate truss-internal tracing instrumentation
    completely from potential user-defined tracing. Opentelemetry has a global state
    that makes "outer" span-contexts parents of nested spans. If user-code in a
    truss model also uses tracing, these traces could easily become polluted with our
    internal contexts. Therefore, all user code (predict and pre/post-processing) should
    be wrapped in this context for isolation.
    """
    current_context = context.get_current()
    # Create an invalid tracing context. This forces that tracing code inside this
    # context manager creates a new root tracing context.
    transient_token = context.attach(trace.set_span_in_context(trace.INVALID_SPAN))
    try:
        yield current_context
    finally:
        # Reattach original context.
        context.detach(transient_token)
        context.attach(current_context)


@contextlib.contextmanager
def section_as_event(span: sdk_trace.Span, section_name: str) -> Iterator[None]:
    """Helper to record the start and end of a sections as events and the duration.

    Note that events are much cheaper to create than dedicated spans.
    """
    t0 = time.time()
    span.add_event(f"start: {section_name}")
    try:
        yield
    finally:
        t1 = time.time()
        span.add_event(
            f"done: {section_name}", attributes={ATTR_NAME_DURATION: t1 - t0}
        )
